import asyncio
import time
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import TimeoutError as FuturesTimeoutError
from contextlib import AsyncExitStack, ExitStack, contextmanager

from dhenara.ai.types import AIModelCallConfig, AIModelCallResponse, AIModelEndpoint
from dhenara.ai.types.genai.dhenara.request import Prompt, SystemInstruction

from .factory import AIModelClientFactory


class AIModelClient:
    """
    A high-level client for making AI model calls with automatic resource management.

    This client handles:
    - Connection lifecycle management
    - Automatic retries with exponential backoff
    - Request timeouts
    - Resource cleanup

    Attributes:
        model_endpoint (AIModelEndpoint): The AI model endpoint configuration
        config (AIModelCallConfig): Configuration for API calls including timeouts and retries
        is_async (bool): Async client or not
    """

    def __init__(
        self,
        model_endpoint: AIModelEndpoint,
        config: AIModelCallConfig | None = None,
        is_async: bool = True,
    ):
        self.model_endpoint = model_endpoint
        self.config = config or AIModelCallConfig()
        self.is_async = is_async
        self._provider_client = None
        self._client_stack = AsyncExitStack() if is_async else ExitStack()

    async def _async_context(self):
        try:
            self._provider_client = await self._client_stack.enter_async_context(
                AIModelClientFactory.create_provider_client(
                    self.model_endpoint,
                    self.config,
                    is_async=True,
                ),
            )
            yield self
        finally:
            await self._client_stack.aclose()
            self._provider_client = None

    @contextmanager
    def _sync_context(self):
        try:
            self._provider_client = self._client_stack.enter_context(
                AIModelClientFactory.create_provider_client(
                    self.model_endpoint,
                    self.config,
                    is_async=False,
                ),
            )
            yield self
        finally:
            self._client_stack.close()
            self._provider_client = None

    def __enter__(self):
        if self.is_async:
            raise RuntimeError("Use 'async with' for async client")
        # Store the context manager instance
        self._sync_ctx = self._sync_context()
        return self._sync_ctx.__enter__()

    def __exit__(self, *exc):
        if self.is_async:
            raise RuntimeError(
                f"This client is created with is_async={self.is_async}. Use 'async with' for async client"
            )
        # Use the stored context manager instance
        return self._sync_ctx.__exit__(*exc)

    async def __aenter__(self):
        if not self.is_async:
            raise RuntimeError(f"This client is created with is_async={self.is_async}. Use 'with' for sync client")
        self._async_ctx = self._async_context()
        return await anext(self._async_ctx.__aiter__())

    async def __aexit__(self, *exc):
        if not self.is_async:
            raise RuntimeError("Use 'with' for sync client")
        try:
            await self._async_ctx.aclose()
        except:
            pass

    def _execute_with_retry_sync(self, *args, **kwargs) -> AIModelCallResponse:
        """Synchronous retry logic"""
        last_exception = None
        for attempt in range(self.config.retries):
            try:
                return self._execute_with_timeout_sync(*args, **kwargs)
            except TimeoutError as e:  # noqa: PERF203 # TODO
                last_exception = e
                if attempt == self.config.retries - 1:
                    break
                delay = min(
                    self.config.retry_delay * (2**attempt),
                    self.config.max_retry_delay,
                )
                time.sleep(delay)

        raise last_exception or RuntimeError("All retry attempts failed")

    async def _execute_with_retry_async(self, *args, **kwargs) -> AIModelCallResponse:
        """Asynchronous retry logic"""
        last_exception = None
        for attempt in range(self.config.retries):
            try:
                return await self._execute_with_timeout_async(*args, **kwargs)
            except asyncio.TimeoutError as e:  # noqa: PERF203 # TODO
                last_exception = e
                if attempt == self.config.retries - 1:
                    break
                delay = min(
                    self.config.retry_delay * (2**attempt),
                    self.config.max_retry_delay,
                )
                await asyncio.sleep(delay)

        raise last_exception or RuntimeError("All retry attempts failed")

    def _execute_with_timeout_sync(self, *args, **kwargs) -> AIModelCallResponse:
        """Synchronous timeout handling using threading instead of signals"""
        if not self.config.timeout:
            # If no timeout is set, just execute directly
            return self._provider_client._format_and_generate_response_sync(*args, **kwargs)

        # Use a thread pool to run the generation function
        with ThreadPoolExecutor(max_workers=1) as executor:
            # Submit the function to the executor
            future = executor.submit(self._provider_client._format_and_generate_response_sync, *args, **kwargs)

            try:
                # Wait for the result with a timeout
                # The timeout value can remain a float here
                return future.result(timeout=self.config.timeout)
            except FuturesTimeoutError:
                # Cancel the future if possible
                future.cancel()
                # Create a custom error message
                raise TimeoutError(f"API call timed out after {self.config.timeout} seconds")

    async def _execute_with_timeout_async(self, *args, **kwargs) -> AIModelCallResponse:
        """Asynchronous timeout handling"""
        async with AsyncExitStack() as stack:
            if self.config.timeout:
                await stack.enter_async_context(asyncio.timeout(self.config.timeout))
            return await self._provider_client._format_and_generate_response_async(*args, **kwargs)

    # Genereate Response Fns
    def generate(
        self,
        prompt: str | dict | Prompt,
        context: list[str | dict | Prompt] | None = None,
        instructions: list[str | dict | SystemInstruction] | None = None,
    ) -> AIModelCallResponse:
        """Synchronous generate method"""
        if self.is_async:
            raise RuntimeError(
                f"This client is created with is_async={self.is_async}. Use generate_async for async client"
            )

        with self as client:  # noqa: F841
            return self._execute_with_retry_sync(
                prompt=prompt,
                context=context,
                instructions=instructions,
            )

    async def generate_async(
        self,
        prompt: str | dict | Prompt,
        context: list[str | dict | Prompt] | None = None,
        instructions: list[str | dict | SystemInstruction] | None = None,
    ) -> AIModelCallResponse:
        """Asynchronous generate method"""
        if not self.is_async:
            raise RuntimeError(f"This client is created with is_async={self.is_async}. Use generate for sync client")

        async with self as client:  # noqa: F841
            return await self._execute_with_retry_async(
                prompt=prompt,
                context=context,
                instructions=instructions,
            )

    async def generate_with_existing_connection(
        self,
        prompt: str | dict | Prompt,
        context: list[str | dict | Prompt] | None = None,
        instructions: list[str | dict | SystemInstruction] | None = None,
    ) -> AIModelCallResponse:
        """
        Generate a response using an existing connection or create a new one.

        This method is useful for making multiple calls efficiently when you want to
        reuse the same connection. Note that you're responsible for cleaning up
        resources when using this method.

        Args:
            prompt: The primary input prompt for the AI model
            context: Optional list of previous conversation context
            instructions: Optional system instructions for the AI model

        Returns:
            AIModelCallResponse: The generated response from the AI model

        Note:
            This method doesn't automatically clean up resources. Use `cleanup()`
            when you're done making calls.
        """
        if self.is_async:
            return self.generate_with_existing_connection_async(
                prompt=prompt, context=context, instructions=instructions
            )
        else:
            return self.generate_with_existing_connection_sync(
                prompt=prompt, context=context, instructions=instructions
            )

    def generate_with_existing_connection_sync(
        self,
        prompt: str | dict | Prompt,
        context: list[str | dict | Prompt] | None = None,
        instructions: list[str | dict | SystemInstruction] | None = None,
    ) -> AIModelCallResponse:
        if not self._provider_client:
            self._provider_client = self._client_stack.enter_context(
                AIModelClientFactory.create_provider_client(
                    self.model_endpoint,
                    self.config,
                    is_async=False,
                ),
            )
        return self._execute_with_retry_sync(
            prompt=prompt,
            context=context,
            instructions=instructions,
        )

    def cleanup_sync(self) -> None:
        """
        Clean up resources manually.

        Call this method when you're done using generate_with_existing_connection()
        to ensure proper resource cleanup.
        """
        self._client_stack.close()
        self._provider_client = None

    async def generate_with_existing_connection_async(
        self,
        prompt: str | dict | Prompt,
        context: list[str | dict | Prompt] | None = None,
        instructions: list[str | dict | SystemInstruction] | None = None,
    ) -> AIModelCallResponse:
        if not self._provider_client:
            self._provider_client = await self._client_stack.enter_async_context(
                AIModelClientFactory.create_provider_client(
                    self.model_endpoint,
                    self.config,
                    is_async=True,
                ),
            )
        return await self._execute_with_retry_async(
            prompt=prompt,
            context=context,
            instructions=instructions,
        )

    async def cleanup_async(self) -> None:
        """
        Clean up resources manually.

        Call this method when you're done using generate_with_existing_connection()
        to ensure proper resource cleanup.
        """
        await self._client_stack.aclose()
        self._provider_client = None



from dhenara.ai.types.shared.base import BaseModel


class AIModelCallResponseMetaData(BaseModel):
    streaming: bool = False
    duration_seconds: int | float | None = None
    provider_metadata: dict | None = None



from typing import Any

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field

from dhenara.ai.types.genai.ai_model import (
    AIModelAPIProviderEnum,
    AIModelProviderEnum,
    ChatResponseUsage,
    UsageCharge,
)
from dhenara.ai.types.genai.dhenara.request import PromptMessageRoleEnum
from dhenara.ai.types.genai.dhenara.request.data import Content, Prompt, PromptConfig, PromptText
from dhenara.ai.types.shared.api import SSEEventType, SSEResponse
from dhenara.ai.types.shared.base import BaseModel

from ._content_items._chat_items import (
    ChatResponseContentItem,
    ChatResponseContentItemDelta,
    ChatResponseContentItemType,
    ChatResponseToolCall,
)
from ._metadata import AIModelCallResponseMetaData


class ChatResponseChoice(BaseModel):
    """A single choice/completion in the chat response"""

    index: int
    finish_reason: Any | None = None
    stop_sequence: Any | None = None
    contents: list[ChatResponseContentItem] | None = None
    metadata: dict = {}

    class Config:
        json_schema_extra = {
            "example": {
                "index": 0,
                "contents": [
                    {
                        "role": "assistant",
                        "text": "Hello! How can I help you today?",
                    }
                ],
            }
        }


class ChatResponseChoiceDelta(BaseModel):
    """A single choice/completion in the chat response"""

    index: int
    finish_reason: Any | None = None
    stop_sequence: Any | None = None
    content_deltas: list[ChatResponseContentItemDelta] | None = None
    metadata: dict = {}


class ChatResponse(BaseModel):
    """Complete chat response from an AI model

    Contains the response content, usage statistics, and provider-specific metadata
    """

    model: str
    provider: AIModelProviderEnum
    api_provider: AIModelAPIProviderEnum | None = None
    usage: ChatResponseUsage | None = None
    usage_charge: UsageCharge | None = None
    choices: list[ChatResponseChoice] = []
    metadata: AIModelCallResponseMetaData | dict = {}

    def get_visible_fields(self) -> dict:
        return self.model_dump(exclude=["choices"])

    def to_prompt(
        self,
        choice_index: int = 0,
        max_words_text: int | None = None,
    ) -> "Prompt":
        """Convert response to a context message for next turn"""

        # Get text from the first choice's contents
        if not self.choices:
            return None

        choice = self.choices[choice_index]
        if not choice.contents:
            return None

        # Combine all content items into one text
        text_parts = [content_item.get_text() for content_item in choice.contents]

        text = "\n".join(text_parts)

        # Create Content object
        content = Content(type="text", text=text)

        # Create PromptText object
        prompt_text = PromptText(content=content)

        # Create and return Prompt object
        return Prompt(
            role=PromptMessageRoleEnum.ASSISTANT,
            text=prompt_text,
            config=PromptConfig(
                max_words_text=max_words_text,
                max_words_file=None,
            ),
        )

    def first(self, content_type: ChatResponseContentItemType):
        "Returns the first content of matching type"
        for choice in self.choices:
            for content in choice.contents:
                if content.type == content_type:
                    return content
        return None

    def text(self) -> str | None:
        "Returns the first text type content"
        text_item = self.first(ChatResponseContentItemType.TEXT)
        return text_item.text if text_item else None

    def reasoning(self) -> str | None:
        "Returns the first thinkning/reasoning type content"
        reasoning_item = self.first(ChatResponseContentItemType.REASONING)
        return reasoning_item.thinking_text if reasoning_item else None

    def tools(self) -> list[ChatResponseToolCall]:
        "Returns all tool type content"
        tools = [
            content
            for choice in self.choices
            for content in choice.contents
            if content.type == ChatResponseContentItemType.TOOL_CALL
        ]
        return tools

    def structured(self) -> dict | None:
        "Returns the first structured-output type content as dict"
        structured_item = self.first(ChatResponseContentItemType.STRUCTURED_OUTPUT)
        return structured_item.structured_output.structured_data if structured_item else None

    def structured_unprocessed(self) -> dict | None:
        "Returns the first structured-output type content as dict"
        structured_item = self.first(ChatResponseContentItemType.STRUCTURED_OUTPUT)
        return structured_item.structured_output

    def structured_pyd(self) -> PydanticBaseModel:
        "Returns the first structured-output type content as its pydantic model instance configured in the input call"
        structured_item = self.first(ChatResponseContentItemType.STRUCTURED_OUTPUT)
        return structured_item.structured_output.as_pydantic() if structured_item else None

    def preview_dict(self):
        """
        Returns a preview version of the response excluding the full content of choices
        but including metadata about them
        """
        _dict = self.model_dump(exclude=["choices"])

        # Add summary information about choices instead of full content
        choice_summaries = []
        for choice in self.choices:
            choice_summary = {
                "index": choice.index,
                "content_count": len(choice.contents),
                "contents_summary": [
                    {
                        "index": content.index,
                        "type": str(content.type),
                    }
                    for content in choice.contents
                ],
            }
            choice_summaries.append(choice_summary)

        _dict["choices_summary"] = choice_summaries
        return _dict


class ChatResponseChunk(BaseModel):
    """Chat response Chunk from an AI model

    Contains the response content, usage statistics, and provider-specific metadata
    """

    model: str
    provider: AIModelProviderEnum
    api_provider: AIModelAPIProviderEnum | None = None
    usage: ChatResponseUsage | None = None
    usage_charge: UsageCharge | None = None
    choice_deltas: list[ChatResponseChoiceDelta] = []
    metadata: AIModelCallResponseMetaData | dict = {}

    done: bool = Field(
        default=False,
        description="Indicates if this is the final chunk",
    )

    def get_visible_fields(self) -> dict:
        return self.model_dump(exclude=["choices"])


class StreamingChatResponse(SSEResponse[ChatResponseChunk]):
    """Specialized SSE response for chat streaming"""

    event: SSEEventType = SSEEventType.TOKEN_STREAM
    data: ChatResponseChunk



from dhenara.ai.types.genai.ai_model import AIModelFunctionalTypeEnum, AIModelProviderEnum

from .anthropic.chat import CHAT_MODELS as ANTHROPIC_CHAT_MODELS
from .deepseek.chat import CHAT_MODELS as DEEPSEEK_CHAT_MODELS
from .google.chat import CHAT_MODELS as GOOGLE_CHAT_MODELS
from .google.image import IMAGE_MODELS as GOOGLE_IMAGE_MODELS
from .openai.chat import CHAT_MODELS as OPENAI_CHAT_MODELS
from .openai.image import IMAGE_MODELS as OPENAI_IMAGE_MODELS

FOUNDATION_MODELS_MAPPINGS = {
    AIModelFunctionalTypeEnum.TEXT_GENERATION: {
        AIModelProviderEnum.OPEN_AI: OPENAI_CHAT_MODELS,
        AIModelProviderEnum.GOOGLE_AI: GOOGLE_CHAT_MODELS,
        AIModelProviderEnum.ANTHROPIC: ANTHROPIC_CHAT_MODELS,
        AIModelProviderEnum.DEEPSEEK: DEEPSEEK_CHAT_MODELS,
    },
    AIModelFunctionalTypeEnum.IMAGE_GENERATION: {
        AIModelProviderEnum.OPEN_AI: OPENAI_IMAGE_MODELS,
        AIModelProviderEnum.GOOGLE_AI: GOOGLE_IMAGE_MODELS,
    },
}

# fmt: off
ALL_CHAT_MODELS = [
    model
    for provider_models in FOUNDATION_MODELS_MAPPINGS[AIModelFunctionalTypeEnum.TEXT_GENERATION].values()
    for model in provider_models
]

ALL_IMAGE_MODELS = [
    model
    for provider_models in FOUNDATION_MODELS_MAPPINGS[AIModelFunctionalTypeEnum.IMAGE_GENERATION].values()
    for model in provider_models
]

ALL_FOUNDATION_MODELS= [
    model
    for functional_type in FOUNDATION_MODELS_MAPPINGS.values()
    for provider_models in functional_type.values()
    for model in provider_models
]
# fmt: on



from ._providers import (
    AIModelAPIProviderEnum as AIModelAPIProviderEnum,
    AIModelFunctionalTypeEnum as AIModelFunctionalTypeEnum,
    AIModelProviderEnum as AIModelProviderEnum,
)
from ._api_provider_configs import PROVIDER_CONFIGS as PROVIDER_CONFIGS
from ._metrics import *

from ._ai_model_api import *
from ._ai_model import *
from ._ai_model_ep import *



from typing import Union

from pydantic import Field

from ._base import BaseResponseContentItem, ChatResponseContentItemType
from ._structured_output import ChatResponseStructuredOutput
from ._tool_call import ChatResponseToolCall


class BaseChatResponseContentItem(BaseResponseContentItem):
    type: ChatResponseContentItemType = Field(
        ...,
        description="Type of the content item",
    )
    role: str | None = Field(
        default=None,
        description="Role of the message sender in the chat context",
    )


class ChatResponseTextContentItem(BaseChatResponseContentItem):
    """Content item specific to chat responses

    Contains the role, text content, and optional function calls for chat interactions

    Attributes:
        role: The role of the message sender (system, user, assistant, or function)
        text: The actual text content of the message
        function_call: Optional function call details if the message involves function calling
    """

    type: ChatResponseContentItemType = ChatResponseContentItemType.TEXT

    text: str | None = Field(
        None,
        description="Plain text content of the message for chat interaction (without reasoning)",
    )

    def get_text(self) -> str:
        return self.text


class ChatResponseReasoningContentItem(BaseChatResponseContentItem):
    type: ChatResponseContentItemType = ChatResponseContentItemType.REASONING

    thinking_text: str | None = Field(
        None,
        description="Thinking text content, for reasoning mode",
    )

    def get_text(self) -> str:
        return self.thinking_text


class ChatResponseToolCallContentItem(BaseChatResponseContentItem):
    type: ChatResponseContentItemType = ChatResponseContentItemType.TOOL_CALL
    tool_call: ChatResponseToolCall = Field(...)

    def get_text(self) -> str:
        if self.tool_call:
            return f"Tool call: {self.tool_call.model_dump()}"
        return str(self.metadata)


class ChatResponseStructuredOutputContentItem(BaseChatResponseContentItem):
    type: ChatResponseContentItemType = ChatResponseContentItemType.STRUCTURED_OUTPUT
    structured_output: ChatResponseStructuredOutput = Field(...)

    def get_text(self) -> str:
        if self.structured_output:
            if self.structured_output.structured_data is not None:
                return f"Structured  Output: {self.structured_output.structured_data}"
            else:
                return f"Structured  Output was failed to parse. Unparsed items: {self.structured_output.model_dump()}"
        return str(self.metadata)


class ChatResponseGenericContentItem(BaseChatResponseContentItem):
    type: ChatResponseContentItemType = ChatResponseContentItemType.GENERIC
    # Use metadata to store the content

    def get_text(self) -> str:
        return str(self.metadata)


ChatResponseContentItem = Union[  # noqa: UP007
    ChatResponseTextContentItem,
    ChatResponseReasoningContentItem,
    ChatResponseToolCallContentItem,
    ChatResponseStructuredOutputContentItem,
    ChatResponseGenericContentItem,
]


# Deltas for streamin
class BaseChatResponseContentItemDelta(BaseResponseContentItem):
    type: ChatResponseContentItemType = Field(
        ...,
        description="Type of the content item",
        serialization_alias="type",  # Ensures type is serialized correctly
    )
    role: str | None = Field(
        default=None,
        description="Role of the message sender in the chat context",
    )


class ChatResponseTextContentItemDelta(BaseChatResponseContentItemDelta):
    type: ChatResponseContentItemType = ChatResponseContentItemType.TEXT

    text_delta: str | None = Field(
        None,
    )

    def get_text_delta(self) -> str:
        return self.text_delta


class ChatResponseReasoningContentItemDelta(BaseChatResponseContentItemDelta):
    type: ChatResponseContentItemType = ChatResponseContentItemType.REASONING

    thinking_text_delta: str | None = Field(
        None,
    )

    def get_text_delta(self) -> str:
        return self.thinking_text_delta


# TODO: Tool call in streaming is not supported now
class ChatResponseToolCallContentItemDelta(BaseChatResponseContentItemDelta):
    type: ChatResponseContentItemType = ChatResponseContentItemType.TOOL_CALL
    tool_calls_delta: str
    tool_call_deltas: list[dict] = Field(default_factory=list)

    def get_text_delta(self) -> str:
        return self.tool_calls_delta


# TODO: Structed output in streaming is not supported now
class ChatResponseStructuredOutputContentItemDelta(BaseChatResponseContentItem):
    pass  # TODO


class ChatResponseGenericContentItemDelta(BaseChatResponseContentItemDelta):
    type: ChatResponseContentItemType = ChatResponseContentItemType.GENERIC
    # Use metadata to store the content

    def get_text_delta(self) -> str:
        return str(self.metadata)


ChatResponseContentItemDelta = Union[  # noqa: UP007
    ChatResponseTextContentItemDelta,
    ChatResponseReasoningContentItemDelta,
    ChatResponseToolCallContentItemDelta,
    ChatResponseStructuredOutputContentItemDelta,
    ChatResponseGenericContentItemDelta,
]



from dhenara.ai.providers.anthropic import AnthropicChat
from dhenara.ai.providers.base import AIModelCallConfig, AIModelProviderClientBase
from dhenara.ai.providers.google import GoogleAIChat, GoogleAIImage
from dhenara.ai.providers.openai import OpenAIChat, OpenAIImage
from dhenara.ai.types.genai.ai_model import AIModelEndpoint, AIModelFunctionalTypeEnum, AIModelProviderEnum


class AIModelClientFactory:
    """Factory for creating AI model providers"""

    _provider_clients = {
        AIModelFunctionalTypeEnum.TEXT_GENERATION: {
            AIModelProviderEnum.OPEN_AI: OpenAIChat,
            AIModelProviderEnum.ANTHROPIC: AnthropicChat,
            AIModelProviderEnum.GOOGLE_AI: GoogleAIChat,
            AIModelProviderEnum.DEEPSEEK: OpenAIChat,
        },
        AIModelFunctionalTypeEnum.IMAGE_GENERATION: {
            AIModelProviderEnum.OPEN_AI: OpenAIImage,
            AIModelProviderEnum.GOOGLE_AI: GoogleAIImage,
        },
    }

    @classmethod
    def create_provider_client(
        cls,
        model_endpoint: AIModelEndpoint,
        config: AIModelCallConfig,
        is_async: bool,
    ) -> AIModelProviderClientBase:
        functional_type_mapping = cls._provider_clients.get(model_endpoint.ai_model.functional_type)
        if not functional_type_mapping:
            raise ValueError(f"Unsupported functional_type: {model_endpoint.ai_model.functional_type}")

        provider_class = functional_type_mapping.get(model_endpoint.ai_model.provider)
        if not provider_class:
            raise ValueError(f"Unsupported provider: {model_endpoint.ai_model.provider}")
        return provider_class(
            model_endpoint=model_endpoint,
            config=config,
            is_async=is_async,
        )



import logging
from typing import Any

from openai.types import ImagesResponse as OpenAIImagesResponse

from dhenara.ai.providers.openai import OpenAIClientBase
from dhenara.ai.types.genai import (
    AIModelCallResponse,
    ImageContentFormat,
    ImageResponse,
    ImageResponseChoice,
    ImageResponseContentItem,
    ImageResponseUsage,
    StreamingChatResponse,
)
from dhenara.ai.types.genai.ai_model import AIModelAPIProviderEnum
from dhenara.ai.types.shared.api import SSEErrorResponse

logger = logging.getLogger(__name__)


class OpenAIImage(OpenAIClientBase):
    """OpenAI Image Generation Client"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Additional params
        self.response_format = None

    def get_api_call_params(
        self,
        prompt: dict,
        context: list[dict] | None = None,
        instructions: dict | None = None,
    ) -> AIModelCallResponse:
        if not self._client:
            raise RuntimeError("Client not initialized. Use with 'async with' context manager")

        if self._input_validation_pending:
            raise ValueError("inputs must be validated with `self.validate_inputs()` before api calls")

        if instructions:
            instructions_str = instructions["content"]
        else:
            instructions_str = ""

        prompt_text = f"{instructions_str} {context} {prompt}"
        model_options = self.model_endpoint.ai_model.get_options_with_defaults(self.config.options)
        user = self.config.get_user()

        image_args: dict[str, Any] = {
            "model": self.model_name_in_api_calls,
            "prompt": prompt_text,
            **model_options,
        }

        if user:
            if self.model_endpoint.api.provider != AIModelAPIProviderEnum.MICROSOFT_AZURE_AI:
                image_args["user"] = user

        # Store additional params
        # NOTE: response_format is not supported for newer models and response is always in b64_json format
        self.response_format = model_options.get("response_format", "b64_json")  # Special case.

        return {"image_args": image_args}

    def do_api_call_sync(
        self,
        api_call_params: dict,
    ) -> AIModelCallResponse:
        image_args = api_call_params["image_args"]
        if self.model_endpoint.api.provider != AIModelAPIProviderEnum.MICROSOFT_AZURE_AI:
            response = self._client.images.generate(**image_args)
        else:
            response = self._client.complete(**image_args)  # Images on Azure NOT tested

        return response

    async def do_api_call_async(
        self,
        api_call_params: dict,
    ) -> AIModelCallResponse:
        image_args = api_call_params["image_args"]
        if self.model_endpoint.api.provider != AIModelAPIProviderEnum.MICROSOFT_AZURE_AI:
            response = await self._client.images.generate(**image_args)
        else:
            response = await self._client.complete(**image_args)  # Images on Azure NOT tested
        return response

    def do_streaming_api_call_sync(
        self,
        api_call_params,
    ) -> AIModelCallResponse:
        raise ValueError("do_streaming_api_call_sync:  Streaming not supported for Image generation")

    async def do_streaming_api_call_async(
        self,
        api_call_params,
    ) -> AIModelCallResponse:
        raise ValueError("do_streaming_api_call_async:  Streaming not supported for Image generation")

    def parse_stream_chunk(
        self,
        chunk,
    ) -> StreamingChatResponse | SSEErrorResponse | None:
        raise ValueError("parse_stream_chunk: Streaming not supported for Image generation")

    def _get_usage_from_provider_response(
        self,
        response: OpenAIImagesResponse,
    ) -> ImageResponseUsage:
        # No usage data availabe in response. We will derive some params
        model = self.model_endpoint.ai_model.model_name
        model_options = self.model_endpoint.ai_model.get_options_with_defaults(self.config.options)

        return ImageResponseUsage(
            number_of_images=len(response.data),
            model=model,
            options=model_options,
        )

    def parse_response(
        self,
        response: OpenAIImagesResponse,
    ) -> ImageResponse:
        """Parse OpenAI image response into standard format"""

        usage, usage_charge = self.get_usage_and_charge(response)

        choices = []
        for idx, image in enumerate(response.data):
            if self.response_format == "b64_json":
                choices.append(
                    ImageResponseChoice(
                        index=idx,
                        contents=[
                            ImageResponseContentItem(
                                index=0,
                                content_format=ImageContentFormat.BASE64,
                                content_b64_json=image.b64_json,
                                metadata={
                                    "revised_prompt": image.revised_prompt,
                                },
                            )
                        ],
                    )
                )

            elif self.response_format == "url":
                choices.append(
                    ImageResponseChoice(
                        index=idx,
                        contents=[
                            ImageResponseContentItem(
                                index=0,
                                content_format=ImageContentFormat.URL,
                                content_url=image.url,
                                metadata={
                                    "revised_prompt": image.revised_prompt,
                                },
                            )
                        ],
                    )
                )
            else:
                raise ValueError(f"Unknown response_format {self.response_format} in parse_response:")

        return ImageResponse(
            model=self.model_endpoint.ai_model.model_name,
            provider=self.model_endpoint.ai_model.provider,
            choices=choices,
            usage=usage,
            usage_charge=usage_charge,
        )



from typing import Any

from dhenara.ai.types.genai.ai_model import AIModelProviderEnum


class ToProviderMixin:
    def to_provider_format(self, provider: AIModelProviderEnum) -> dict[str, Any]:
        """Convert to provider format"""
        if provider == AIModelProviderEnum.OPEN_AI:
            return self.to_openai_format()
        elif provider == AIModelProviderEnum.ANTHROPIC:
            return self.to_anthropic_format()
        elif provider == AIModelProviderEnum.GOOGLE_AI:
            return self.to_google_format()
        else:
            raise ValueError(f"Provider {provider} not supported")


class FromProviderMixin:
    def from_provider_format(self, provider: AIModelProviderEnum) -> dict[str, Any]:
        """Convert to provider format"""
        if provider == AIModelProviderEnum.OPEN_AI:
            return self.from_openai_format()
        elif provider == AIModelProviderEnum.ANTHROPIC:
            return self.from_anthropic_format()
        elif provider == AIModelProviderEnum.GOOGLE_AI:
            return self.from_google_format()
        else:
            raise ValueError(f"Provider {provider} not supported")



from dhenara.ai.types.genai.ai_model import (
    AIModelFunctionalTypeEnum,
    AIModelProviderEnum,
    FoundationModel,
    ImageModelCostData,
    ImageModelSettings,
    ValidOptionValue,
)

DallE2 = FoundationModel(
    model_name="dall-e-2",
    display_name="dall-e-2",
    provider=AIModelProviderEnum.OPEN_AI,
    functional_type=AIModelFunctionalTypeEnum.IMAGE_GENERATION,
    settings=ImageModelSettings(max_words=1000),
    valid_options={
        "size": ValidOptionValue(
            allowed_values=["256x256", "512x512", "1024x1024"],
            default_value="1024x1024",
            cost_sensitive=True,
            description="Dimensions",
        ),
        "n": ValidOptionValue(
            allowed_values=list(range(1, 10)),
            default_value=1,
            cost_sensitive=True,
            description="Number of images",
        ),
        "response_format": ValidOptionValue(
            allowed_values=["b64_json", "url"],
            default_value="url",
            cost_sensitive=False,
            description="Response format",
        ),
    },
    metadata={
        "details": "DALL·E model released in Nov 2023.",
    },
    cost_data=ImageModelCostData(
        flat_cost_per_image=None,  # Pricing is within options
        image_options_cost_data=[
            {
                "size": ["256x256"],
                "cost_per_image": 0.016,
            },
            {
                "size": ["512x512"],
                "cost_per_image": 0.018,
            },
            {
                "size": ["1024x1024"],
                "cost_per_image": 0.020,
            },
        ],
    ),
)


DallE3 = FoundationModel(
    model_name="dall-e-3",
    display_name="dall-e-3",
    provider=AIModelProviderEnum.OPEN_AI,
    functional_type=AIModelFunctionalTypeEnum.IMAGE_GENERATION,
    settings=ImageModelSettings(max_words=4000),
    valid_options={
        "quality": ValidOptionValue(
            allowed_values=["standard", "hd"],
            default_value="standard",
            cost_sensitive=True,
            description="Image quality",
        ),
        "size": ValidOptionValue(
            allowed_values=["1024x1024", "1024x1792", "1792x1024"],
            default_value="1024x1024",
            cost_sensitive=True,
            description="Dimensions",
        ),
        "style": ValidOptionValue(
            allowed_values=["natural", "vivid"],
            default_value="natural",
            cost_sensitive=False,
            description="Style preference",
        ),
        "n": ValidOptionValue(
            allowed_values=[1],
            default_value=1,
            cost_sensitive=True,
            description="Number of images",
        ),
        "response_format": ValidOptionValue(
            allowed_values=["b64_json", "url"],
            default_value="url",
            cost_sensitive=False,
            description="Response format",
        ),
    },
    metadata={
        "details": "DALL·E model released in Nov 2023.",
    },
    cost_data=ImageModelCostData(
        flat_cost_per_image=None,  # Pricing is within options
        image_options_cost_data=[
            {
                "quality": ["standard"],
                "size": ["1024x1024"],
                "cost_per_image": 0.04,
            },
            {
                "quality": ["standard"],
                "size": ["1024x1792", "1792x1024"],
                "cost_per_image": 0.08,
            },
            {
                "quality": ["hd"],
                "size": ["1024x1024"],
                "cost_per_image": 0.08,
            },
            {
                "quality": ["hd"],
                "size": ["1024x1792", "1792x1024"],
                "cost_per_image": 0.12,
            },
        ],
    ),
)


GPTImage1 = FoundationModel(
    model_name="gpt-image-1",
    display_name="gpt-image-1",
    provider=AIModelProviderEnum.OPEN_AI,
    functional_type=AIModelFunctionalTypeEnum.IMAGE_GENERATION,
    settings=ImageModelSettings(max_words=1000),
    valid_options={
        "quality": ValidOptionValue(
            allowed_values=["low", "medium", "high"],
            default_value="medium",
            cost_sensitive=True,
            description="Image quality",
        ),
        "size": ValidOptionValue(
            allowed_values=["1024x1024", "1024x1536", "1536x1024"],
            default_value="1024x1024",
            cost_sensitive=True,
            description="Dimensions",
        ),
        "n": ValidOptionValue(
            allowed_values=list(range(1, 10)),
            default_value=1,
            cost_sensitive=True,
            description="Number of images",
        ),
        # NO "response_format", responses will be always in b64_json
    },
    metadata={
        "details": "GPTImage 1",
    },
    cost_data=ImageModelCostData(
        flat_cost_per_image=None,  # Pricing is within options
        image_options_cost_data=[
            {
                "quality": ["low"],
                "size": ["1024x1024"],
                "cost_per_image": 0.011,
            },
            {
                "quality": ["low"],
                "size": ["1024x1536", "1536x1024"],
                "cost_per_image": 0.016,
            },
            {
                "quality": ["medium"],
                "size": ["1024x1024"],
                "cost_per_image": 0.042,
            },
            {
                "quality": ["medium"],
                "size": ["1024x1536", "1536x1024"],
                "cost_per_image": 0.063,
            },
            {
                "quality": ["high"],
                "size": ["1024x1024"],
                "cost_per_image": 0.167,
            },
            {
                "quality": ["high"],
                "size": ["1024x1536", "1536x1024"],
                "cost_per_image": 0.25,
            },
        ],
    ),
)


IMAGE_MODELS = [DallE2, DallE3, GPTImage1]



import base64

from pydantic import Field

from dhenara.ai.types.shared.base import BaseEnum, BaseModel

from ._base import BaseResponseContentItem


class ImageContentFormat(BaseEnum):
    """Enum representing different formats of image content"""

    URL = "url"
    BASE64 = "base64"
    BYTES = "bytes"
    UNKNOWN = "unknown"


class ImageResponseContentItem(BaseResponseContentItem):
    """Content item specific to image generation responses

    Contains the generated image data in various formats (bytes, base64, or URL)

    Attributes:
        content_bytes: Raw image bytes
        content_b64_json: Base64 encoded image data
        content_url: URL to the generated image
        format: Image format (e.g., PNG, JPEG)
        size: Image dimensions
    """

    content_format: ImageContentFormat = Field(
        ...,
        description="Response content format",
    )
    content_bytes: bytes | None = Field(
        None,
        description="Raw image content in bytes",
    )
    content_b64_json: str | None = Field(
        None,
        description="Base64 encoded image content",
        min_length=1,
    )
    content_url: str | None = Field(
        None,
        description="URL to access the generated image",
        pattern=r"^https?://.*$",
    )

    def validate_content(self) -> bool:
        """Validates that at least one content field is populated

        Returns:
            bool: True if at least one content field has data
        """
        return any(
            [
                self.content_bytes is not None,
                self.content_b64_json is not None,
                self.content_url is not None,
            ]
        )

    def get_content_as_bytes(self) -> bytes:
        if self.content_format == ImageContentFormat.BYTES:
            byte_content = self.content_bytes
        elif self.content_format == ImageContentFormat.BASE64:
            byte_content = base64.b64decode(self.content_b64_json)
        else:
            raise ValueError(
                f"get_content_as_bytes: Content format {self.content_format} not supported."
                "Only byte and b64_json is supported now"
            )

        return byte_content


class UsageCharge(BaseModel):
    cost: float = Field(
        ...,
        description="Cost",
    )
    charge: float | None = Field(
        ...,
        description="Charge after considering internal expences and margins."
        " Will be  None if  `cost_multiplier_percentage` is not set in cost data.",
    )



import logging

from google.oauth2 import service_account

from dhenara.ai.types.genai.ai_model import AIModelAPI, AIModelAPIProviderEnum

logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
class APIProviderSharedFns:
    # -------------------------------------------------------------------------
    # client params for vertext ai is diffente
    @staticmethod
    def get_vertex_ai_credentials(
        api: AIModelAPI,
    ):
        if api.provider == AIModelAPIProviderEnum.GOOGLE_VERTEX_AI:
            client_params = api.get_provider_credentials()

            service_account_json = client_params["service_account_json"]
            project_id = client_params["project_id"]
            location = client_params["location"]

            scopes = [
                "https://www.googleapis.com/auth/cloud-platform",
                "https://www.googleapis.com/auth/cloud-platform.read-only",
            ]

            sa_credentials = service_account.Credentials.from_service_account_info(service_account_json, scopes=scopes)

            #  Vertex AI API
            final_client_params = {
                "credentials": sa_credentials,
                "project_id": project_id,
                "location": location,
            }
            return final_client_params
        else:
            logger.error(
                f"get_vertex_ai_client_params should only be called for api with provider vertext ai. "
                f"provider={api.provider}"
            )
            return {}



from pydantic import Field

from dhenara.ai.types.shared.base import BaseEnum, BaseModel


class BaseResponseContentItem(BaseModel):
    """Base content item for AI model responses

    Contains common metadata fields used across different types of AI responses

    Attributes:
        metadata: System-generated metadata from API response
        storage_metadata: Storage-related metadata (e.g., cloud storage information)
        custom_metadata: User-defined additional metadata
    """

    index: int = Field(
        default=0,
        description="Content item index",
    )
    metadata: dict = Field(
        default_factory=dict,
        description="System-generated metadata from API response processing",
    )
    storage_metadata: dict = Field(
        default_factory=dict,
        description=(
            "User-defined storage-related metadata such as cloud storage details, paths, or references. "
            "Will be empty on output from `dhenara-ai` package."
        ),
    )
    custom_metadata: dict = Field(
        default_factory=dict,
        description=(
            "User-defined additional metadata for custom processing or tracking."
            "Will be empty on output from `dhenara-ai` package"
        ),
    )


class ChatResponseContentItemType(BaseEnum):
    """Enum representing different types of content items in chat responses"""

    TEXT = "text"
    REASONING = "reasoning"
    GENERIC = "generic"
    TOOL_CALL = "tool_call"
    STRUCTURED_OUTPUT = "structured_output"



from datetime import datetime
from typing import Annotated

from pydantic import PlainSerializer

# custom datetime type for serialization to ISO format
ISODateTime = Annotated[
    datetime,
    PlainSerializer(lambda dt: dt.isoformat(), return_type=str),
]



import json
from typing import Any

from pydantic import Field

from dhenara.ai.types.shared.base import BaseModel


class ChatResponseToolCall(BaseModel):
    """Representation of a tool call from an LLM"""

    id: str | None = None
    name: str
    arguments: dict[str, Any] = Field(default_factory=dict)
    raw_data: str | dict | None = Field(
        None,
        description="Raw unparsed response from the model",
    )
    parse_error: str | None = Field(
        None,
        description="Error that occurred during parsing, if any",
    )

    @classmethod
    def _parse(cls, arguments: str | dict) -> dict:
        """Parse arguments from either JSON striing or dict"""
        arguments_dict = {}
        raw_data = None
        parse_error = None
        if isinstance(arguments, str):
            try:
                arguments_dict = json.loads(arguments)
            except Exception as e:
                raw_data = arguments
                parse_error = str(e)
        elif isinstance(arguments, dict):
            try:
                arguments_dict = arguments
            except Exception as e:
                raw_data = arguments
                parse_error = str(e)
        else:
            raw_data = arguments
            parse_error = f"Invalid arguments type {(type(arguments))}"

        return {
            "arguments_dict": arguments_dict,
            "raw_data": raw_data,
            "parse_error": parse_error,
        }

    @classmethod
    def from_openai_format(cls, props: dict) -> "ChatResponseToolCall":
        """Create from OpenAI tool call format"""
        _args = props.get("function", {}).get("arguments")
        _parse_result = cls._parse(_args)

        return cls(
            id=props.get("id"),
            name=props.get("function", {}).get("name"),
            arguments=_parse_result.get("arguments_dict"),
            raw_data=_parse_result.get("raw_data"),
            parse_error=_parse_result.get("parse_error"),
        )

    @classmethod
    def from_anthropic_format(cls, props: dict) -> "ChatResponseToolCall":
        """Create from Anthropic tool use format"""
        _args = props.get("input")
        _parse_result = cls._parse(_args)

        return cls(
            id=props.get("id"),
            name=props.get("name"),
            arguments=_parse_result.get("arguments_dict"),
            raw_data=_parse_result.get("raw_data"),
            parse_error=_parse_result.get("parse_error"),
        )

    @classmethod
    def from_google_format(cls, props: dict) -> "ChatResponseToolCall":
        _args = props.get("args")
        _parse_result = cls._parse(_args)

        return cls(
            id=props.get("id"),
            name=props.get("name"),
            arguments=_parse_result.get("arguments_dict"),
            raw_data=_parse_result.get("raw_data"),
            parse_error=_parse_result.get("parse_error"),
        )


class ChatResponseToolCallResult(BaseModel):
    """Result of executing a tool call, which may pass to LLM in next turn"""

    tool_name: str
    call_id: str | None = None
    result: Any = None
    error: str | None = None

    def to_openai_format(self) -> dict:
        """Convert to OpenAI format for tool response"""
        return {
            "tool_call_id": self.call_id,
            "role": "tool",
            "name": self.tool_name,
            "content": json.dumps(self.result) if self.result is not None else str(self.error),
        }

    def to_anthropic_format(self) -> dict:
        """Convert to Anthropic format for tool response"""
        return {
            "type": "tool_result",
            "tool_use_id": self.call_id,
            "content": json.dumps(self.result) if self.result is not None else str(self.error),
        }



import logging
from typing import Literal

from pydantic import Field

from dhenara.ai.types.shared.base import BaseModel
from dhenara.ai.types.shared.file import FileContentData, FileContentFormatEnum, FileFormatEnum, FileMetadata

logger = logging.getLogger(__name__)


# TODO: Need cleanup
# -----------------------------------------------------------------------------
class GenericFileContent(BaseModel):
    content_format: Literal["bytes", "base64"] = Field(
        default=None,
        description="Content format",
    )
    content_bytes: bytes | None = Field(
        None,
        description="Raw image content in bytes",
    )
    content_b64_json: str | None = Field(
        None,
        description="Base64 encoded image content",
    )


class GenericFile(BaseModel):
    name: str
    metadata: FileMetadata | None = None


# -----------------------------------------------------------------------------
class StoredFile(GenericFile):
    url: str | None = Field(default=None)
    content: GenericFileContent | None = Field(
        default=None,
        description="File Content ",
    )
    path: str | None = Field(default=None)


# -----------------------------------------------------------------------------
class ProcessedFile(GenericFile):
    processed_content: FileContentData

    def get_source_file_name(self) -> str:
        """Get the original source file name"""
        return self.processed_content.name

    def get_file_format(self) -> FileFormatEnum:
        """Get the file format enum"""
        return self.processed_content.file_format

    def get_content_format(self) -> FileContentFormatEnum:
        """Get the content format enum"""
        return self.processed_content.content_format

    def get_metadata(self) -> FileMetadata | None:
        """Get file metadata"""
        # TODO_FUTURE : for processed files, metatdata shalll be duplicated in the content and in the file metadata
        return self.processed_content.metadata

    def get_mime_type(self) -> str | None:
        """Get the mime type of the file"""
        mime_type = self.get_metadata()
        if mime_type:
            return mime_type.mime_type.lower()
        return None

    def get_processed_file_data(self, max_words: int | None = None) -> str:
        """Get processed file data with optional word limit"""

        if max_words == 0:
            return ""

        # content = str(self.processed_content.content or "")
        # content = self.model_dump_json()
        _content = self.model_dump()
        content = str(_content)

        if max_words is not None:
            words = content.split()
            content = " ".join(words[:max_words])

        return content

    def get_processed_file_data_content_only(self):
        if self.processed_content:
            # NOTE: Its impossilbe to return `only` the content in case of zip files.
            # Here we will return the full content.
            # This should be OK as long as zip files won't need to be send as image/byte in context. TODO_FUTURE
            if self.processed_content.contents:
                return self.get_processed_file_data()

            content_str = self.processed_content.content
            return content_str
        else:
            return ""



from pydantic import Field

from dhenara.ai.types.genai.ai_model import AIModelAPIProviderEnum
from dhenara.ai.types.shared.base import BaseEnum, BaseModel


class ExternalApiCallStatusEnum(BaseEnum):
    """Status codes for AI model API endpoint responses."""

    REQUEST_SEND = "request_send"
    REQUEST_NOT_SEND = "request_not_send"
    RESPONSE_RECEIVED_SUCCESS = "response_received_success"
    RESPONSE_RECEIVED_API_ERROR = "response_received_api_error"
    RESPONSE_TIMEOUT = "response_timeout"
    INTERNAL_PROCESSING_ERROR = "internal_processing_error"


class ExternalApiCallStatus(BaseModel):
    """
    Status information for an AI model API call.

    Tracks the request/response status, provider details, and any error information
    for calls made to AI model APIs.
    """

    status: ExternalApiCallStatusEnum = Field(
        ...,
        description="Current status of the API call",
    )
    api_provider: AIModelAPIProviderEnum | str = Field(
        ...,
        description="Name of the AI API provider (e.g. 'OpenAI', 'Anthropic')",
        min_length=1,
    )
    model: str = Field(
        ...,
        description="Name of the specific AI model used",
        min_length=1,
    )
    message: str = Field(
        ...,
        description="Human readable status/error message",
    )
    code: str | None = Field(
        None,
        description="Provider-specific status/error code",
    )
    http_status_code: int | None = Field(
        None,
        description="HTTP status code if applicable",
        ge=100,
        le=599,
    )
    data: dict | None = Field(
        None,
        description="Additional provider-specific status data",
    )

    @property
    def successful(self) -> bool:
        """Check if the API call completed successfully."""
        return self.status == ExternalApiCallStatusEnum.RESPONSE_RECEIVED_SUCCESS

    class Config:
        json_schema_extra = {
            "example": {
                "status": "response_received_success",
                "api_provider": "OpenAI",
                "model": "gpt-4",
                "message": "Request completed successfully",
                "code": "200",
                "http_status_code": 200,
                "data": {"usage": {"total_tokens": 150}},
            }
        }



import logging
import re
from typing import Any

from openai.types.chat import ChatCompletion, ChatCompletionMessage
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk, ChoiceDelta

from dhenara.ai.providers.openai import OpenAIClientBase
from dhenara.ai.types.genai import (
    AIModelCallResponse,
    AIModelCallResponseMetaData,
    ChatResponse,
    ChatResponseChoice,
    ChatResponseChoiceDelta,
    ChatResponseContentItem,
    ChatResponseContentItemDelta,
    ChatResponseReasoningContentItem,
    ChatResponseReasoningContentItemDelta,
    ChatResponseStructuredOutput,
    ChatResponseStructuredOutputContentItem,
    ChatResponseTextContentItem,
    ChatResponseTextContentItemDelta,
    ChatResponseToolCall,
    ChatResponseToolCallContentItem,
    ChatResponseUsage,
    StreamingChatResponse,
)
from dhenara.ai.types.genai.ai_model import AIModelAPIProviderEnum, AIModelProviderEnum
from dhenara.ai.types.shared.api import SSEErrorResponse

logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
class OpenAIChat(OpenAIClientBase):
    def get_api_call_params(
        self,
        prompt: dict,
        context: list[dict] | None = None,
        instructions: dict | None = None,
    ) -> AIModelCallResponse:
        if not self._client:
            raise RuntimeError("Client not initialized. Use with 'async with' context manager")

        if self._input_validation_pending:
            raise ValueError("inputs must be validated with `self.validate_inputs()` before api calls")

        messages = []
        user = self.config.get_user()

        # Process instructions
        if instructions:
            messages.append(instructions)

        # Add previous messages and current prompt
        if context:
            messages.extend(context)

        messages.append(prompt)

        # Prepare API call arguments
        chat_args: dict[str, Any] = {
            "model": self.model_name_in_api_calls,
            "messages": messages,
            "stream": self.config.streaming,
        }

        if user:
            if self.model_endpoint.api.provider != AIModelAPIProviderEnum.MICROSOFT_AZURE_AI:
                chat_args["user"] = user

        max_output_tokens, max_reasoning_tokens = self.config.get_max_output_tokens(self.model_endpoint.ai_model)

        if max_output_tokens is not None:
            if self.model_endpoint.api.provider != AIModelAPIProviderEnum.MICROSOFT_AZURE_AI:
                # NOTE: With resasoning models, max_output_tokens Deprecated in favour of max_completion_tokens
                chat_args["max_completion_tokens"] = max_output_tokens

            else:
                chat_args["max_tokens"] = max_output_tokens

        if self.config.streaming:
            if self.model_endpoint.api.provider != AIModelAPIProviderEnum.MICROSOFT_AZURE_AI:
                chat_args["stream_options"] = {"include_usage": True}

        if self.config.options:
            chat_args.update(self.config.options)

        # ---  Tools ---
        if self.config.tools:
            chat_args["tools"] = self.formatter.format_tools(
                tools=self.config.tools,
                model_endpoint=self.model_endpoint,
            )

        if self.config.tool_choice:
            chat_args["tool_choice"] = self.formatter.format_tool_choice(
                tool_choice=self.config.tool_choice,
                model_endpoint=self.model_endpoint,
            )

        # --- Structured Output ---
        if self.config.structured_output:
            chat_args["response_format"] = self.formatter.format_structured_output(
                structured_output=self.config.structured_output,
                model_endpoint=self.model_endpoint,
            )

        return {"chat_args": chat_args}

    def do_api_call_sync(
        self,
        api_call_params: dict,
    ) -> AIModelCallResponse:
        chat_args = api_call_params["chat_args"]
        if self.model_endpoint.api.provider != AIModelAPIProviderEnum.MICROSOFT_AZURE_AI:
            response = self._client.chat.completions.create(**chat_args)
        else:
            response = self._client.complete(**chat_args)
        return response

    async def do_api_call_async(
        self,
        api_call_params: dict,
    ) -> AIModelCallResponse:
        chat_args = api_call_params["chat_args"]
        if self.model_endpoint.api.provider != AIModelAPIProviderEnum.MICROSOFT_AZURE_AI:
            response = await self._client.chat.completions.create(**chat_args)
        else:
            response = await self._client.complete(**chat_args)
        return response

    def do_streaming_api_call_sync(
        self,
        api_call_params,
    ) -> AIModelCallResponse:
        chat_args = api_call_params["chat_args"]
        if self.model_endpoint.api.provider != AIModelAPIProviderEnum.MICROSOFT_AZURE_AI:
            stream = self._client.chat.completions.create(**chat_args)
        else:
            stream = self._client.complete(**chat_args)

        return stream

    async def do_streaming_api_call_async(
        self,
        api_call_params,
    ) -> AIModelCallResponse:
        chat_args = api_call_params["chat_args"]
        if self.model_endpoint.api.provider != AIModelAPIProviderEnum.MICROSOFT_AZURE_AI:
            stream = await self._client.chat.completions.create(**chat_args)
        else:
            stream = await self._client.complete(**chat_args)

        return stream

    # -------------------------------------------------------------------------
    def parse_stream_chunk(
        self,
        chunk: ChatCompletionChunk,
    ) -> StreamingChatResponse | SSEErrorResponse | None:
        """Handle streaming response with progress tracking and final response"""
        processed_chunks = []

        self.streaming_manager.provider_metadata = None
        self.streaming_manager.persistant_choice_metadata_list = []

        if not self.streaming_manager.provider_metadata:  # Grab the metadata once
            self.streaming_manager.provider_metadata = {
                "id": chunk.id,
                "created": str(chunk.created),  # Microsoft sdk returns datetim obj
                "object": chunk.object if hasattr(chunk, "object") else None,
                "system_fingerprint": chunk.system_fingerprint if hasattr(chunk, "system_fingerprint") else None,
            }

        # Process usage
        if hasattr(chunk, "usage") and chunk.usage:  # Microsoft is slow in adopting openai changes 😶
            usage = ChatResponseUsage(
                total_tokens=chunk.usage.total_tokens,
                prompt_tokens=chunk.usage.prompt_tokens,
                completion_tokens=chunk.usage.completion_tokens,
            )
            self.streaming_manager.update_usage(usage)

        # Process content
        if chunk.choices:
            choice_deltas = []
            for choice in chunk.choices:
                # Only first chunk has few fields
                if len(self.streaming_manager.persistant_choice_metadata_list) < choice.index + 1:
                    self.streaming_manager.persistant_choice_metadata_list.append(
                        {
                            "role": choice.delta.role,
                            "refusal": choice.delta.refusal if hasattr(choice.delta, "refusal") else None,
                        }
                    )

                role = choice.delta.role or self.streaming_manager.persistant_choice_metadata_list[choice.index]["role"]

                choice_deltas.append(
                    ChatResponseChoiceDelta(
                        index=choice.index,
                        finish_reason=choice.finish_reason if hasattr(choice, "finish_reason") else None,
                        stop_sequence=None,
                        content_deltas=[
                            self.process_content_item_delta(
                                index=0,  # Only one content item. Might change with reasoing response?
                                role=role,
                                delta=choice.delta,
                            ),
                        ],
                        metadata={},
                    )
                )

            response_chunk = self.streaming_manager.update(choice_deltas=choice_deltas)
            stream_response = StreamingChatResponse(
                id=chunk.id,
                data=response_chunk,
            )

            processed_chunks.append(stream_response)

        return processed_chunks

    # -------------------------------------------------------------------------
    def _get_usage_from_provider_response(
        self,
        response: ChatCompletion,
    ) -> ChatResponseUsage:
        return ChatResponseUsage(
            total_tokens=response.usage.total_tokens,
            prompt_tokens=response.usage.prompt_tokens,
            completion_tokens=response.usage.completion_tokens,
        )

    # -------------------------------------------------------------------------
    def parse_response(
        self,
        response: ChatCompletion,
    ) -> ChatResponse:
        """Parse the OpenAI response into our standard format"""
        usage, usage_charge = self.get_usage_and_charge(response)

        choices = []
        for choice in response.choices:
            content_items = self.process_content_item(
                index=0,  # Only one content item. Might change with reasoing response?
                role=choice.message.role,
                content_item=choice.message,
            )

            contents = content_items if isinstance(content_items, list) else [content_items]

            choice_obj = ChatResponseChoice(
                index=choice.index,
                finish_reason=choice.finish_reason if hasattr(choice, "finish_reason") else None,
                stop_sequence=None,
                contents=contents,
                metadata={},
            )
            choices.append(choice_obj)

        return ChatResponse(
            model=response.model,
            provider=self.model_endpoint.ai_model.provider,
            api_provider=self.model_endpoint.api.provider,
            usage=usage,
            usage_charge=usage_charge,
            choices=choices,
            metadata=AIModelCallResponseMetaData(
                streaming=False,
                duration_seconds=0,  # TODO
                provider_metadata={
                    "id": response.id,
                    "created": str(response.created),  # Microsoft sdk returns datetim obj
                    "object": response.object if hasattr(response, "object") else None,
                    # TODO: Move to choice
                    "system_fingerprint": response.system_fingerprint
                    if hasattr(response, "system_fingerprint")
                    else None,
                },
            ),
        )

    def process_content_item(
        self,
        index: int,
        role: str,
        content_item: ChatCompletionMessage,
    ) -> ChatResponseContentItem | list[ChatResponseContentItem]:
        # INFO: response type will vary with API/Model providers.
        # if isinstance(content_item, (ChatCompletionMessage, AzureChatResponseMessage)):

        if hasattr(content_item, "content") and content_item.content:
            all_items = []
            _content = content_item.content
            _index = index

            if self.model_endpoint.ai_model.provider == AIModelProviderEnum.DEEPSEEK:
                answer_content = None
                reasoning_content = None
                # Extract reasoning and answer content
                think_match = re.search(r"<think>(.*?)</think>", _content, re.DOTALL)
                if think_match:
                    reasoning_content = think_match.group(1).strip()
                    answer_content = re.sub(r"<think>.*?</think>", "", _content, flags=re.DOTALL).strip()

                if reasoning_content:
                    _index = index + 1
                    _content = answer_content
                    all_items.append(
                        ChatResponseReasoningContentItem(
                            index=index,
                            role=role,
                            thinking_text=reasoning_content,
                        )
                    )

            if self.config.structured_output is not None:
                structured_output = ChatResponseStructuredOutput.from_model_output(
                    raw_response=_content,
                    config=self.config.structured_output,
                )
                all_items.append(
                    ChatResponseStructuredOutputContentItem(
                        index=_index,
                        role=role,
                        structured_output=structured_output,
                    )
                )
            else:
                all_items.append(
                    ChatResponseTextContentItem(
                        index=_index,
                        role=role,
                        text=_content,
                    )
                )

            return all_items

        elif hasattr(content_item, "tool_calls") and content_item.tool_calls:
            tool_call_items = []
            for tool_call in content_item.tool_calls:
                if isinstance(tool_call, dict):
                    content_item_dict = tool_call
                else:
                    content_item_dict = tool_call.model_dump()

                tool_call_items.append(
                    ChatResponseToolCallContentItem(
                        index=index,
                        role=role,
                        tool_call=ChatResponseToolCall.from_openai_format(content_item_dict),
                        metadata={},
                    )
                )
            return tool_call_items
        else:
            return self.get_unknown_content_type_item(
                index=index,
                role=role,
                unknown_item=content_item,
                streaming=False,
            )

    # Streaming
    def process_content_item_delta(
        self,
        index: int,
        role: str,
        delta: ChoiceDelta,
    ) -> ChatResponseContentItemDelta:
        # if isinstance(delta, (ChoiceDelta, AzureStreamingChatResponseMessageUpdate)):
        if hasattr(delta, "content"):
            if self.model_endpoint.ai_model.provider == AIModelProviderEnum.DEEPSEEK:
                content = delta.content

                # Check for think tag markers
                think_start = "<think>" in content
                think_end = "</think>" in content

                # If we see a start tag, everything after it goes to reasoning
                if think_start:
                    self.streaming_manager.progress.in_thinking_block = True
                    # Split content at the tag
                    _, reasoning_part = content.split("<think>", 1)
                    return ChatResponseReasoningContentItemDelta(
                        index=index,
                        role=role,
                        thinking_text_delta=reasoning_part,
                    )

                # If we see an end tag, everything before it goes to reasoning
                elif think_end:
                    self.streaming_manager.progress.in_thinking_block = False
                    reasoning_part, answer_part = content.split("</think>", 1)
                    return (
                        ChatResponseReasoningContentItemDelta(
                            index=index,
                            role=role,
                            thinking_text_delta=reasoning_part,
                        )
                        if reasoning_part
                        else ChatResponseTextContentItemDelta(
                            index=index + 1,
                            role=role,
                            text_delta=answer_part,
                        )
                    )

                # If we're inside a thinking block, content goes to reasoning
                elif self.streaming_manager.progress.in_thinking_block:
                    return ChatResponseReasoningContentItemDelta(
                        index=index,
                        role=role,
                        thinking_text_delta=content,
                    )

                # Otherwise it's regular text content
                else:
                    return ChatResponseTextContentItemDelta(
                        index=index + 1,
                        role=role,
                        text_delta=content,
                    )
            else:
                return ChatResponseTextContentItemDelta(
                    index=index,
                    role=role,
                    text_delta=delta.content,
                )
        # TODO: Tools Not supported in streaming yet
        # elif hasattr(delta, "tool_calls") and delta.tool_calls:
        #    tool_call_deltas = []

        #    for tool_call in delta.tool_calls:
        #        tool_call_delta = {
        #            "id": tool_call.id if hasattr(tool_call, "id") else None,
        #            "type": "function",  # OpenAI currently only supports function type
        #            "function": {},
        #        }

        #        # Handle function name
        #        if hasattr(tool_call, "function") and hasattr(tool_call.function, "name"):
        #            tool_call_delta["function"]["name"] = tool_call.function.name

        #        # Handle function arguments (can be streamed)
        #        if hasattr(tool_call, "function") and hasattr(tool_call.function, "arguments"):
        #            tool_call_delta["function"]["arguments"] = tool_call.function.arguments

        #        tool_call_deltas.append(tool_call_delta)

        #    return ChatResponseToolCallContentItemDelta(
        #        index=index,
        #        role=role,
        #        tool_call_deltas=tool_call_deltas,
        #        metadata={},
        #    )

        else:
            # NOTE: There is no way to identify content type for OpenAI streaming response.
            # Also they sends few extra chuckw with no content.
            # Eg: The very first chunk will only have the `role` set (and role in other chunks will be None)
            # Therefore, dont't send a `ChatResponseGenericContentItemDelta`  here,
            # asit will messup streaming manager content updation

            return self.get_unknown_content_type_item(
                index=index,
                role=role,
                unknown_item=delta,
                streaming=True,
            )



import json
from typing import Any

# from urllib.request import urlopen
from pydantic import Field

from dhenara.ai.types.shared.base import BaseEnum, BaseModel


class ContentType(BaseEnum):
    """Enumeration of content types that can be returned."""

    TEXT = "text"
    LIST = "list"  # Text List
    DICT = "dict"  # Json
    JSONL = "jsonl"


class Content(BaseModel):
    """Represents user input data for AI model processing.

    This model handles various forms of input content including text, URLs, and JSON data
    that can be processed by AI models.

    """

    type: ContentType = Field(
        default=None,
    )
    text: str | None = Field(
        default=None,
        description="Primary text content to be processed",
        example="What is the capital of France?",
    )
    textl: list[str] | None = Field(
        default=None,
        description="Multiple text contents for batch processing",
        example=["Text 1", "Text 2"],
    )

    # Not using name `json` as it shadows an attribute in parent "BaseModel"
    json_c: dict[str, Any] = Field(
        default_factory=dict,
        description="Structured JSON data for processing",
        example={"key": "value"},
    )
    jsonl_c: list[dict[str, Any]] = Field(
        default_factory=list,
        description="List of JSON objects in JSONL format",
        example=[{"id": 1, "text": "example"}, {"id": 2, "text": "example2"}],
    )

    class Config:
        """Pydantic model configuration."""

        json_schema_extra = {
            "examples": [
                {
                    "text": "What is the capital of France?",
                    "textl": ["Text 1", "Text 2"],
                    "json_c": {"key": "value"},
                    "jsonl_c": [{"id": 1, "text": "example"}],
                },
            ],
        }

    @property
    def has_content(self) -> bool:
        """Check if the input contains any content.

        Returns:
            bool: True if any content field is populated, False otherwise.
        """
        return any(
            [
                self.text is not None,
                self.textl is not None,
                bool(self.json_c),
                bool(self.jsonl_c),
            ]
        )

    # def validate_content_length(self, content: str) -> str:
    #    if len(content) > 8192:
    #        raise ValueError("Content exceeds maximum length of 8192 characters")
    #    return content

    def get_content(
        self,
        separator: str = "\n",
    ) -> str | list[str] | dict | list[dict]:
        """Retrieve content from any of the input sources.

        This method consolidates content from various input fields and returns it in the
        requested format. It handles text content, URLs, JSON, and JSONL data.

        Args:
            return_type: Desired return type format (text, list, dict, or jsonl_c)
            separator: String separator for joining text content when return_type is TEXT

        Returns:
            Content in the requested format:
            - TEXT: Single string with all content joined
            - LIST: List of strings
            - DICT: Dictionary from json_c
            - JSONL: List of dictionaries from jsonl_c
        """
        if not self.has_content:
            raise ValueError("No content available in any field")

        # Collect all text content
        text_list: list[str] = []

        # Add single content if present
        if self.text:
            text_list.append(self.text)

        # Add multiple textl if present
        if self.textl:
            text_list.extend(self.textl)

        # Return based on requested type
        if self.type == ContentType.TEXT:
            # Include JSON content if present
            if self.json_c:
                text_list.append(json.dumps(self.json_c, ensure_ascii=False))
            if self.jsonl_c:
                text_list.extend(json.dumps(item, ensure_ascii=False) for item in self.jsonl_c)
            return separator.join(text_list)

        elif self.type == ContentType.LIST:
            return text_list

        elif self.type == ContentType.DICT:
            if not self.json_c:
                raise ValueError("No JSON content available")
            return self.json_c

        elif self.type == ContentType.JSONL:
            if not self.jsonl_c:
                raise ValueError("No JSONL content available")
            return self.jsonl_c
        else:
            raise ValueError(f"get_content: Unknown type {self.type}")



import json
import logging
from typing import Any, TypeVar

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field

from dhenara.ai.types.genai.dhenara import StructuredOutputConfig
from dhenara.ai.types.shared.base import BaseModel

from ._tool_call import ChatResponseToolCall

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=PydanticBaseModel)


def _coerce_json_strings(obj: Any) -> Any:
    """
    Recursively process data structures and convert JSON strings to Python objects.

    If a string looks like JSON (starts with { or [ and ends with } or ]),
    attempt to parse it into a Python object.
    """
    if isinstance(obj, str):
        s = obj.strip()
        if (s.startswith("{") and s.endswith("}")) or (s.startswith("[") and s.endswith("]")):
            try:
                return _coerce_json_strings(json.loads(s))
            except json.JSONDecodeError:
                # Not valid JSON, return the original string
                return obj
        return obj
    elif isinstance(obj, list):
        return [_coerce_json_strings(v) for v in obj]
    elif isinstance(obj, dict):
        return {k: _coerce_json_strings(v) for k, v in obj.items()}
    else:
        return obj


class ChatResponseStructuredOutput(BaseModel):
    """Content item specific to structured output responses

    Contains the structured data output from the model according to a specified schema

    Attributes:
        type: The type of content item (always STRUCTURED_OUTPUT)
        structured_data: The parsed structured data
        raw_data: The raw unparsed response from the model
        schema: The schema that was used for the structured output
        parse_error: Any error that occurred during parsing
    """

    config: StructuredOutputConfig = Field(
        ...,
        description="StructuredOutputConfig used for generating this response",
    )
    structured_data: dict | None = Field(
        None,
        description="Parsed structured data according to the schema",
    )
    raw_data: str | dict | None = Field(
        None,
        description="Raw unparsed response from the model",
    )
    parse_error: str | None = Field(
        None,
        description="Error that occurred during parsing, if any",
    )

    def get_text(self) -> str:
        """Get a text representation of the structured data"""
        if self.structured_data is not None:
            return str(self.structured_data)
        elif self.raw_data is not None:
            return str(self.raw_data)
        elif self.parse_error is not None:
            return f"Error parsing structured output: {self.parse_error}"
        return ""

    def as_pydantic(
        self,
        model_class: type[PydanticBaseModel] | None = None,
    ) -> PydanticBaseModel | None:
        """Convert the structured data to a pydantic model instance

        Args:
            model_class: Optional pydantic model class to use for conversion.
                         If not provided, uses the original schema class if available.

        Returns:
            Pydantic model instance or None if conversion fails
        """
        if self.structured_data is None:
            return None

        if not model_class:
            model_class = self.config.model_class_reference

        try:
            if model_class is not None:
                return model_class.model_validate(self.structured_data)
            else:
                logger.error("Error: need model_class to convert to pydantic model")
                return None
        except Exception as e:
            logger.error(f"Error converting structured data to pydantic model: {e}")
            return None

    @classmethod
    def _parse_and_validate(
        cls,
        raw_data: str | dict,
        config: StructuredOutputConfig,
    ) -> tuple[dict | None, str | None]:
        """Parse and validate data against schema, handling nested JSON strings"""
        error = None
        parsed_data = None

        try:
            # Step 1: Initial parsing if the input is a string
            initial_data = raw_data
            if isinstance(raw_data, str):
                try:
                    initial_data = json.loads(raw_data)
                except json.JSONDecodeError:
                    # Not valid JSON at top level, keep as string for model validation
                    initial_data = raw_data

            # Step 2: Recursively normalize all nested JSON strings
            normalized_data = _coerce_json_strings(initial_data)

            # Step 3: Get model class from config
            model_cls: type[PydanticBaseModel] = None
            if isinstance(config.model_class_reference, type) and issubclass(
                config.model_class_reference, PydanticBaseModel
            ):
                model_cls = config.model_class_reference
            elif isinstance(config.model_class_reference, PydanticBaseModel):
                model_cls = config.model_class_reference.__class__

            # Step 4: Validate with model class if available
            if model_cls:
                try:
                    # Always use model_validate instead of model_validate_json since
                    # we've already parsed any JSON strings
                    parsed_data_pyd = model_cls.model_validate(normalized_data)
                    parsed_data = parsed_data_pyd.model_dump()
                except Exception as e:
                    logger.exception(f"Model validation error: {e}")
                    error = str(e)
            else:
                # No model class available, just return the normalized data
                parsed_data = normalized_data

        except Exception as e:
            logger.exception(f"Unexpected error during parsing/validation: {e}")
            error = str(e)

        return parsed_data, error

    @classmethod
    def from_model_output(
        cls,
        raw_response: str | dict,
        config: StructuredOutputConfig,
    ) -> "ChatResponseStructuredOutput":
        """Create a structured output content item from model output

        Args:
            raw_response: Raw output from the model
            config: Schema to validate against

        Returns:
            ChatResponseStructuredOutput
        """

        raw_response_to_parse = raw_response or {}
        parsed_data, error = cls._parse_and_validate(raw_response_to_parse, config)

        return cls(
            config=config,
            structured_data=parsed_data,
            raw_data=raw_response,  # Keep original response regardless of parsing
            parse_error=error,
        )

    @classmethod
    def from_tool_call(
        cls,
        raw_response: str | dict,
        tool_call: ChatResponseToolCall | None,
        config: StructuredOutputConfig,
    ) -> "ChatResponseStructuredOutput":
        """Create a structured output from a tool call response

        Args:
            tool_call: The tool call response
            config: StructuredOutputConfig to use for validation

        Returns:
            ChatResponseStructuredOutput instance
        """

        if tool_call is not None:
            if tool_call.arguments:
                raw_response_to_parse = tool_call.arguments  # Get the dict directly
                parsed_data, error = cls._parse_and_validate(raw_response_to_parse, config)
                # In case of error, keep the  orginal data
                raw_data = raw_response if error is not None else None
            else:
                parsed_data = None
                error = tool_call.parse_error
                raw_data = raw_response
        else:
            parsed_data = None
            error = "No tool call provided with `from_tool_call` method"
            raw_data = raw_response

        return cls(
            config=config,
            structured_data=parsed_data,
            raw_data=raw_data,  # Keep original response regardless of parsing
            parse_error=error,
        )



import logging

from google.genai.types import GenerateImagesConfig, GenerateImagesResponse

from dhenara.ai.providers.google import GoogleAIClientBase
from dhenara.ai.types.genai import (
    AIModelCallResponse,
    ImageContentFormat,
    ImageResponse,
    ImageResponseChoice,
    ImageResponseContentItem,
    ImageResponseUsage,
)

logger = logging.getLogger(__name__)


class GoogleAIImage(GoogleAIClientBase):
    """GoogleAI Image Generation Client"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_api_call_params(
        self,
        prompt: dict,
        context: list[dict] | None = None,
        instructions: dict | None = None,
    ) -> AIModelCallResponse:
        if not self._client:
            raise RuntimeError("Client not initialized. Use with 'async with' context manager")

        if self._input_validation_pending:
            raise ValueError("Inputs must be validated before API calls")

        if instructions:
            instructions_str = instructions["parts"][0]["text"]
        else:
            instructions_str = ""

        prompt_text = f"{instructions_str} {context} {prompt}"

        generate_config_args = self.get_default_generate_config_args()
        generate_config = GenerateImagesConfig(**generate_config_args)

        return {
            "prompt": prompt_text,
            "generate_config": generate_config,
        }

    def do_api_call_sync(
        self,
        api_call_params: dict,
    ) -> AIModelCallResponse:
        response = self._client.models.generate_images(
            model=self.model_name_in_api_calls,
            config=api_call_params["generate_config"],
            prompt=api_call_params["prompt"],
        )
        return response

    async def do_api_call_async(
        self,
        api_call_params: dict,
    ) -> AIModelCallResponse:
        response = await self._client.models.generate_images(
            model=self.model_name_in_api_calls,
            config=api_call_params["generate_config"],
            prompt=api_call_params["prompt"],
        )
        return response

    def do_streaming_api_call_sync(
        self,
        api_call_params,
    ) -> AIModelCallResponse:
        raise ValueError("do_streaming_api_call_sync:  Streaming not supported for Image generation")

    async def do_streaming_api_call_async(
        self,
        api_call_params,
    ) -> AIModelCallResponse:
        raise ValueError("do_streaming_api_call_async:  Streaming not supported for Image generation")

    def get_default_generate_config_args(self) -> dict:
        model_options = self.model_endpoint.ai_model.get_options_with_defaults(self.config.options)

        config_params = {
            **model_options,
            "output_mime_type": "image/jpeg",
            "include_rai_reason": True,
            "safety_filter_level": "block_only_high",
            # "person_generation": "allow_adult",
            # "negative_prompt":"Outside","human"
        }

        return config_params

    def parse_stream_chunk(
        self,
        chunk,
    ):
        raise ValueError("parse_stream_chunk: Streaming not supported for Image generation")

    def _get_usage_from_provider_response(
        self,
        response: GenerateImagesResponse,
    ) -> ImageResponseUsage:
        # No usage data availabe in response. We will derive some params
        model = self.model_endpoint.ai_model.model_name
        model_options = self.model_endpoint.ai_model.get_options_with_defaults(self.config.options)

        return ImageResponseUsage(
            number_of_images=len(response.generated_images),
            model=model,
            options=model_options,
        )

    def parse_response(
        self,
        response: GenerateImagesResponse,
    ) -> ImageResponse:
        """Parse GoogleAI image response into standard format"""

        usage, usage_charge = self.get_usage_and_charge(response)
        choices = []
        for idx, image in enumerate(response.generated_images):
            choices.append(
                ImageResponseChoice(
                    index=idx,
                    contents=[
                        ImageResponseContentItem(
                            index=0,
                            content_format=ImageContentFormat.BYTES,
                            content_bytes=image.image.image_bytes,
                            metadata={
                                "rai_filtered_reason": image.rai_filtered_reason,
                            },
                        )
                    ],
                )
            )

        return ImageResponse(
            model=self.model_endpoint.ai_model.model_name,
            provider=self.model_endpoint.ai_model.provider,
            choices=choices,
            usage=usage,
            usage_charge=usage_charge,
        )



import inspect
from collections.abc import Callable
from typing import Any, Literal, get_type_hints

from pydantic import BaseModel, Field, create_model

from dhenara.ai.types.genai.dhenara.request import ToProviderMixin


class FunctionParameter(BaseModel, ToProviderMixin):
    """Parameter definition for function/tool parameters"""

    type: str = Field(..., description="Type of the parameter (string, number, boolean, etc.)")
    description: str | None = Field(default=None, description="Description of the parameter")
    required: bool = Field(default=False, description="Whether the parameter is required")
    allowed_values: list[Any] | None = Field(default=None, description="Allowed values")
    default: Any | None = Field(default=None, description="Default value for the parameter")


class FunctionParameters(BaseModel, ToProviderMixin):
    """Schema for function parameters"""

    type: Literal["object"] = "object"
    properties: dict[str, FunctionParameter] = Field(..., description="Properties of the function parameters")
    required: list[str] | None = Field(default_factory=list, description="List of required parameters")


class FunctionDefinition(BaseModel, ToProviderMixin):
    """Generic function/tool definition that works across all providers"""

    name: str = Field(..., description="Name of the function")
    description: str | None = Field(default=None, description="Description of the function")
    parameters: FunctionParameters = Field(..., description="Parameters for the function")


class ToolDefinition(BaseModel, ToProviderMixin):
    """Tool definition that wraps a function"""

    type: Literal["function"] = "function"
    function: FunctionDefinition = Field(..., description="Function definition")
    function_reference: Callable | None = Field(default=None, description="Callable fn reference")

    @classmethod
    def from_callable(cls, func: Callable) -> "ToolDefinition":
        """
        Create a ToolDefinition from a Python callable.

        Args:
            func: A Python function to convert to a tool definition

        Returns:
            A ToolDefinition object representing the function
        """
        # Get function signature and docstring
        signature = inspect.signature(func)
        doc = inspect.getdoc(func) or ""

        # Get type hints
        type_hints = get_type_hints(func)

        # Create field definitions for Pydantic model
        fields = {}
        required_params = []

        for name, param in signature.parameters.items():
            if name == "self":
                continue

            # Get type annotation
            annotation = type_hints.get(name, Any)

            # Get parameter description
            param_desc = None
            for line in doc.split("\n"):
                if f":param {name}:" in line:
                    param_desc = line.split(f":param {name}:")[1].strip()
                    break

            # Determine if parameter is required
            is_required = param.default == inspect.Parameter.empty
            default = ... if is_required else param.default

            if is_required:
                required_params.append(name)

            # Add field to model
            fields[name] = (annotation, Field(default=default, description=param_desc))

        # Create a Pydantic model dynamically
        _params_model = create_model(f"{func.__name__}Params", **fields)

        # Get JSON schema from Pydantic model
        schema = _params_model.model_json_schema()

        # Convert to FunctionParameters
        properties = {}
        for prop_name, prop_schema in schema.get("properties", {}).items():
            json_type = prop_schema.get("type", "string")
            description = prop_schema.get("description")

            properties[prop_name] = FunctionParameter(
                type=json_type,
                description=description,
                required=prop_name in required_params,
                default=None if prop_name in required_params else signature.parameters[prop_name].default,
                allowed_values=prop_schema.get("enum"),
            )

        # Create function definition
        function_def = FunctionDefinition(
            name=func.__name__,
            description=doc.split("\n\n")[0] if doc else None,  # First paragraph of docstring
            parameters=FunctionParameters(properties=properties, required=required_params),
        )

        # Create and return tool definition
        return cls(function=function_def, function_reference=func)


# TODO: Add privider specific fns
class BuiltInTool(BaseModel):
    """Built-in Tool by provider"""

    type: str
    name: str


class ToolChoice(BaseModel):
    type: Literal["zero_or_more", "one_or_more", "specific"] | None = Field(
        default="zero_or_more",
        description=(
            "Tool choice type. "
            "NOTE: A `None` will make disable all tolls and responses will be as if like without a toll"
        ),
    )
    specific_tool_name: str | None = None



from pydantic import Field

from dhenara.ai.types.shared.base import BaseModel


class UsageCharge(BaseModel):
    cost: float = Field(
        ...,
        description="Cost",
    )
    charge: float | None = Field(
        default=None,
        description="Charge after considering internal expences and margins."
        " Will be  None if  `cost_multiplier_percentage` is not set in cost data.",
    )


class ChatResponseUsage(BaseModel):
    """Token usage statistics for the chat completion"""

    total_tokens: int
    prompt_tokens: int
    completion_tokens: int

    class Config:
        json_schema_extra = {
            "example": {
                "total_tokens": 100,
                "prompt_tokens": 50,
                "completion_tokens": 50,
            }
        }


class ImageResponseUsage(BaseModel):
    """Usage information for image generation.
    Note that, for images, no usage data is received, so this class holds params required for usage/cost calculation"""

    number_of_images: int = Field(
        ...,
        description="Number of Images generated",
    )
    model: str = Field(
        default_factory=dict,
        description="Model Name",
    )
    options: dict = Field(
        default_factory=dict,
        description="Options send to API",
    )

    class Config:
        json_schema_extra = {
            "example": {
                "model": "dall-e-3",
                "options": {
                    "size": "1024x1024",
                    "quality": "standard",
                },
            }
        }



from typing import Any

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field

from dhenara.ai.types.shared.base import BaseModel


class StructuredOutputConfig(BaseModel):
    """Configuration for structured output"""

    # This field will be used for serialization
    output_schema: dict[str, Any] = Field(
        ...,
        description="Schema for the structured output as a JSON schema dictionary",
    )

    # This field will need to be included in the model but marked as excluded from serialization
    model_class_reference: type[PydanticBaseModel] | None = Field(
        None,
        description="Reference to the Pydantic model class (not serialized)",
        exclude=True,  # This ensures it's excluded from serialization
    )

    @classmethod
    def from_model(cls, model_class: type[PydanticBaseModel]):
        """Create config from a model class"""
        return cls(
            output_schema=model_class.model_json_schema(),
            model_class_reference=model_class,  # Store the reference for programmatic access
        )

    def get_schema(self) -> dict[str, Any]:
        """Get the schema object"""
        return self.output_schema

    def get_model_class(self) -> type[PydanticBaseModel] | None:
        """Get the model class for Python operations"""
        return self.model_class_reference

    # Override model_dump to handle the model_class field
    def model_dump(self, **kwargs):
        # Get the default serialization
        data = super().model_dump(**kwargs)
        # Remove model_class if it's in the output
        if "model_class_reference" in data:
            del data["model_class_reference"]
        return data



import logging

from pydantic import BaseModel as PydanticBaseModel
from pydantic import model_validator

from dhenara.ai.types.genai.ai_model import AIModel
from dhenara.ai.types.genai.dhenara.request import (
    StructuredOutputConfig,
    ToolChoice,
    ToolDefinition,
)
from dhenara.ai.types.shared.base import BaseModel

logger = logging.getLogger(__name__)


# TODO_FUTURE:
# Create seperate class for the parameter required conversion,
# class AIModelCallData(BaseModel):
#    tools: list[ToolDefinition] | None = None
#    tool_choice: ToolChoice | None = None
#    structured_output: StructuredOutputConfig | None = None


class AIModelCallConfig(BaseModel):
    """Configuration for AI model calls"""

    streaming: bool = False
    max_output_tokens: int | None = None
    reasoning: bool = False
    max_reasoning_tokens: int | None = None
    options: dict = {}

    tools: list[ToolDefinition] | None = None
    tool_choice: ToolChoice | None = None

    structured_output: type[PydanticBaseModel] | StructuredOutputConfig | None = None

    metadata: dict = {}
    timeout: float | None = None
    retries: int = 3
    retry_delay: float = 1.0
    max_retry_delay: float = 10.0
    test_mode: bool = False

    @model_validator(mode="after")
    def validate_structured_output(self) -> "AIModelCallConfig":
        if isinstance(self.structured_output, type) and issubclass(self.structured_output, PydanticBaseModel):
            self.structured_output = StructuredOutputConfig.from_model(
                model_class=self.structured_output,
            )
        return self

    def get_user(self):
        user = self.metadata.get("user", None)
        if not user:
            user = self.metadata.get("user_id", None)

        return user

    def get_max_output_tokens(self, model: AIModel) -> tuple[int, int | None]:
        """Returns max_output_tokens and max_reasoning_tokens based on the model settings and call-config"""

        if not model:
            raise ValueError("Model should be passed when max_token is not set in the call-config")

        _settings = model.get_settings()

        # Determine which max output tokens to use based on reasoning mode
        if not self.reasoning:
            _settings_max_output_tokens = _settings.max_output_tokens
            _reasoning_capable = False
        elif not _settings.supports_reasoning:  # Don't flag an error
            _settings_max_output_tokens = _settings.max_output_tokens
            _reasoning_capable = False
        else:
            _settings_max_output_tokens = _settings.max_output_tokens_reasoning_mode
            _reasoning_capable = True

        if not _settings_max_output_tokens:
            token_type = "max_output_tokens_reasoning_mode" if _reasoning_capable else "max_output_tokens"
            raise ValueError(f"Invalid call-config. {token_type} is not set in model {model.model_name}.")

        # Set max output tokens
        max_output_tokens = min(
            self.max_output_tokens if self.max_output_tokens is not None else _settings_max_output_tokens,
            _settings_max_output_tokens,
        )

        # Set max reasoning tokens
        if not _reasoning_capable or not self.reasoning or _settings.max_reasoning_tokens is None:
            max_reasoning_tokens = None
        else:
            max_reasoning_tokens = min(
                self.max_reasoning_tokens if self.max_reasoning_tokens is not None else _settings.max_reasoning_tokens,
                _settings.max_reasoning_tokens,
            )

        return (max_output_tokens, max_reasoning_tokens)



import json
from typing import Any

from pydantic import Field, field_validator, model_validator

from dhenara.ai.types.genai.ai_model import PROVIDER_CONFIGS, AIModelAPIProviderEnum
from dhenara.ai.types.shared.base import BaseModel


class AIModelAPI(BaseModel):
    """
    Pydantic model representing API credentials for AI model providers with built-in validation.
    """

    provider: AIModelAPIProviderEnum = Field(
        ...,
        description="The AI model provider",
    )
    api_key: str | None = Field(
        None,
        description="API key, if applicable ",
    )
    credentials: dict[str, Any] | None = Field(
        None,
        description="Dictionary of sensitive credentials otherthan API key.",
    )
    config: dict[str, Any] = Field(
        default_factory=dict,
        description="Non-sensitive credential configuration/parameters.",
    )
    order: int = Field(
        0,
        description="Order for display purposes",
    )
    enabled: bool = Field(
        True,  # noqa: FBT003
        description="Whether this API is enabled",
    )
    reference_number: str | None = Field(
        None,
        description="Reference number. Should be unique if not None",
    )

    @model_validator(mode="after")
    def validate_provider_requirements(self) -> "AIModelAPI":
        """Validate provider-specific requirements"""
        provider_config = PROVIDER_CONFIGS.get(self.provider)
        if not provider_config:
            raise ValueError(f"Unsupported provider: {self.provider}")

        # Validate API key requirement
        if provider_config.api_key_required and not self.api_key:
            raise ValueError(f"API key is required for {self.provider}")

        # Validate required credentials
        if self.credentials:
            for field_config in provider_config.credentials_required_fields:
                value = self.credentials.get(field_config.field_name)
                if not value:
                    raise ValueError(f"Missing required credential: {field_config.field_name}")

                if field_config.is_json_field:
                    try:
                        if isinstance(value, dict):
                            parsed_value = value
                        else:
                            parsed_value = json.loads(value.strip())

                        if not isinstance(parsed_value, dict):
                            raise ValueError(f"JSON field {field_config.field_name} must be a dictionary")

                    except json.JSONDecodeError as e:
                        raise ValueError(f"Invalid JSON in {field_config.field_name}: {e}")

        # Validate required config fields
        for field_config in provider_config.config_required_fields:
            if field_config.field_name not in self.config:
                raise ValueError(f"Missing required config: {field_config.field_name}")

        return self

    @field_validator("api_key")
    @classmethod
    def validate_api_key(cls, v: str | None) -> str | None:
        """Validate API key format"""
        if v is not None and len(v) < 8:
            raise ValueError("API key must be at least 8 characters long")
        return v

    @field_validator("credentials")
    @classmethod
    def validate_credentials(cls, v: dict[str, Any] | None) -> dict[str, Any] | None:
        if v is None:
            return v

        # Create a copy of the dictionary
        validated = v.copy()

        # Handle JSON fields if needed
        for key, value in v.items():
            if isinstance(value, str) and key.endswith("_json"):  # TODO: Use get_credentials_fields_config_with_json()
                try:
                    # Remove leading/trailing whitespace and newlines
                    validated[key] = json.loads(value.strip())
                except json.JSONDecodeError as e:
                    raise ValueError(f"Invalid JSON in {key}: {e}")

        return validated

    def get_provider_credentials(self) -> dict[str, Any]:
        """Get provider-specific credentials based on output mappings"""
        provider_config = PROVIDER_CONFIGS.get(self.provider)
        if not provider_config:
            raise ValueError(f"Unsupported provider: {self.provider}")

        result = {}
        for mapping in provider_config.output_mappings:
            value = None
            if mapping.source == "api_key":
                value = self.api_key
            elif mapping.source == "credentials" and mapping.source_key:
                value = self.credentials.get(mapping.source_key) if self.credentials else None
            elif mapping.source == "config" and mapping.source_key:
                value = self.config.get(mapping.source_key)

            if value is None and mapping.default_value is not None:
                value = mapping.default_value

            result[mapping.output_key] = value

        return result

    @classmethod
    def get_providers_form_config(cls) -> dict[str, dict]:
        """
        Get form configuration for all providers in a JSON-serializable format.

        Returns:
            Dictionary mapping provider names to their form configurations.
        """
        providers_config = {}

        for provider, config in PROVIDER_CONFIGS.items():
            credentials_fields = []
            config_fields = []

            # Process required credential fields
            for field in config.credentials_required_fields:
                credentials_fields.append(  # noqa: PERF401
                    {
                        "field_name": field.field_name,
                        "label": field.field_name.replace("_", " ").title(),
                        "type": "textarea" if field.field_name == "service_account_json" else "text",
                        "required": True,
                        "error_msg": field.error_msg,
                    }
                )

            # Process optional credential fields
            for field in config.credentials_optional_fields:
                credentials_fields.append(  # noqa: PERF401
                    {
                        "field_name": field.field_name,
                        "label": field.field_name.replace("_", " ").title(),
                        "type": "text",
                        "required": False,
                        "error_msg": field.error_msg,
                    }
                )

            # Process required config fields
            for field in config.config_required_fields:
                config_fields.append(  # noqa: PERF401
                    {
                        "field_name": field.field_name,
                        "label": field.field_name.replace("_", " ").title(),
                        "type": "text",
                        "required": True,
                        "error_msg": field.error_msg,
                    }
                )

            # Process optional config fields
            for field in config.config_optional_fields:
                config_fields.append(  # noqa: PERF401
                    {
                        "field_name": field.field_name,
                        "label": field.field_name.replace("_", " ").title(),
                        "type": "text",
                        "required": False,
                        "error_msg": field.error_msg,
                    }
                )

            providers_config[provider.value] = {
                "value": provider.value,
                "api_key_required": config.api_key_required,
                "credentials_fields": credentials_fields,
                "config_fields": config_fields,
            }

        return providers_config

    def get_masked_credentials(self) -> dict[str, Any]:
        """Get masked version of credentials for display"""
        masked = {}
        if self.api_key:
            masked["api_key"] = f"{self.api_key[:4]}...{self.api_key[-4:]}"
        if self.credentials:
            masked["credentials"] = {
                k: f"{str(v)[:4]}...{str(v)[-4:]}" if v else None for k, v in self.credentials.items()
            }
        if self.config:
            masked["config"] = self.config
        return masked

    @classmethod
    def validate_updates(
        cls,
        current_model: "AIModelAPI",
        api_key: str | None = None,
        credentials: dict[str, Any] | None = None,
        config: dict[str, Any] | None = None,
    ) -> "AIModelAPI":
        """
        Validate updates to an existing model configuration while preserving all other fields.

        Args:
            current_model: Current AIModelAPI instance
            api_key: New API key to update
            credentials: New credentials to merge
            config: New config to merge

        Returns:
            AIModelAPI: New validated model instance with updates

        Raises:
            ValueError: If updates are invalid
        """
        # Start with current model's data
        current_data = current_model.model_dump()

        # Update only the fields that are provided
        updated_data = current_data.copy()

        if api_key is not None:
            updated_data["api_key"] = api_key

        if credentials is not None:
            updated_data["credentials"] = {
                **current_data.get("credentials", {}),
                **credentials,
            }

        if config is not None:
            updated_data["config"] = {
                **current_data.get("config", {}),
                **config,
            }

        # Create and validate new model with all fields
        try:
            updated_model = cls(**updated_data)
            return updated_model
        except Exception as e:
            raise ValueError(f"Invalid updates: {e}") from e

    @classmethod
    def validate_credentials_update(
        cls,
        current_model: "AIModelAPI",
        credentials: dict[str, Any],
    ) -> dict[str, Any]:
        """Validate credentials update only."""
        return cls.validate_updates(
            current_model,
            credentials=credentials,
        ).credentials

    @classmethod
    def validate_config_update(
        cls,
        current_model: "AIModelAPI",
        config: dict[str, Any],
    ) -> dict[str, Any]:
        """Validate config update only."""
        return cls.validate_updates(
            current_model,
            config=config,
        ).config

    # *************************************************************************
    # Override representation for masking sensitive data
    # TODO_FUTURE: Find better alternative
    def __str__(self) -> str:
        return self._get_masked_representation()

    def __repr__(self) -> str:
        return self._get_masked_representation()

    # *************************************************************************



import logging
from typing import Any

from dhenara.ai.providers.base import BaseFormatter
from dhenara.ai.types.genai.ai_model import AIModelEndpoint, AIModelFunctionalTypeEnum
from dhenara.ai.types.genai.dhenara.request import (
    FunctionDefinition,
    FunctionParameter,
    FunctionParameters,
    PromptMessageRoleEnum,
    StructuredOutputConfig,
    ToolChoice,
    ToolDefinition,
)
from dhenara.ai.types.genai.dhenara.request.data import FormattedPrompt
from dhenara.ai.types.shared.file import FileFormatEnum, GenericFile

logger = logging.getLogger(__name__)


class OpenAIFormatter(BaseFormatter):
    """
    Formatter for converting Dhenara types to OpenAI-specific formats and vice versa.
    """

    role_map = {
        PromptMessageRoleEnum.USER: "user",
        PromptMessageRoleEnum.ASSISTANT: "assistant",
        PromptMessageRoleEnum.SYSTEM: "system",
    }

    @classmethod
    def convert_prompt(
        cls,
        formatted_prompt: FormattedPrompt,
        model_endpoint: AIModelEndpoint | None = None,
        files: list[GenericFile] | None = None,
        max_words_file: int | None = None,
    ) -> dict[str, Any]:
        # Map Dhenara formats to provider format
        file_contents = None
        if files:
            file_contents = cls.convert_files_to_provider_content(
                files=files,
                model_endpoint=model_endpoint,
                max_words=max_words_file,
            )

        if model_endpoint.ai_model.functional_type == AIModelFunctionalTypeEnum.IMAGE_GENERATION:
            return cls._convert_image_model_prompt(
                formatted_prompt=formatted_prompt,
                model_endpoint=model_endpoint,
                file_contents=file_contents,
            )

        if file_contents:
            content = [
                {
                    "type": "text",
                    "text": formatted_prompt.text,
                },
                *file_contents,
            ]
        else:
            # Use Simple text formtat
            content = formatted_prompt.text

        role = cls.role_map.get(formatted_prompt.role)
        return {"role": role, "content": content}

    @classmethod
    def convert_instruction_prompt(
        cls,
        formatted_prompt: FormattedPrompt,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        # Beta models won't support System role
        if model_endpoint.ai_model.beta:
            role = cls.role_map.get(PromptMessageRoleEnum.USER)
        else:
            role = cls.role_map.get(formatted_prompt.role)

        return {"role": role, "content": formatted_prompt.text}

    @classmethod
    def convert_files_to_provider_content(
        cls,
        files: list[GenericFile],
        model_endpoint: AIModelEndpoint | None = None,
        max_words: int | None = None,
    ) -> list[dict[str, Any]]:
        if model_endpoint.ai_model.functional_type == AIModelFunctionalTypeEnum.IMAGE_GENERATION:
            return cls._convert_files_for_image_models(
                files=files,
                model_endpoint=model_endpoint,
                max_words=max_words,
            )

        contents = []
        for file in files:
            file_format = file.get_file_format()
            try:
                if file_format in [FileFormatEnum.COMPRESSED, FileFormatEnum.TEXT]:
                    text = f"\nFile: {file.get_source_file_name()}  Content: {file.get_processed_file_data(max_words)}"
                    pcontent = {
                        "type": "text",
                        "text": text,
                    }
                    contents.append(pcontent)
                elif file_format in [FileFormatEnum.IMAGE]:
                    mime_type = file.get_mime_type()
                    if mime_type not in ["image/jpeg", "image/png", "image/gif", "image/webp"]:
                        raise ValueError(f"Unsupported media type: {mime_type} for file {file.name}")

                    data_content = file.get_processed_file_data_content_only()
                    pcontent = {
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:{mime_type};base64,{data_content}",
                        },
                    }
                    contents.append(pcontent)
                else:
                    logger.error(f"convert_file_content: Unknown file_format {file_format} for file {file.name} ")
            except Exception as e:
                logger.error(f"Error processing file {file.name}: {e}")

        return contents

    # -------------------------------------------------------------------------
    # Internal helper fn
    @classmethod
    def _convert_image_model_prompt(
        cls,
        formatted_prompt: FormattedPrompt,
        file_contents: list[dict[str, Any]],
        model_endpoint: AIModelEndpoint | None = None,
    ) -> str:
        if file_contents:
            _file_content = " ".join(file_contents)
            content = formatted_prompt.text + " " + _file_content
        else:
            content = formatted_prompt.text

        return content

    @classmethod
    def _convert_files_for_image_models(
        cls,
        files: list[GenericFile],
        model_endpoint: AIModelEndpoint | None = None,
        max_words: int | None = None,
    ) -> str:
        contents: list[dict[str, Any]] = []
        for file in files:
            file_format = file.get_file_format()
            try:
                if file_format in [FileFormatEnum.COMPRESSED, FileFormatEnum.TEXT]:
                    text = (
                        f"\nFile: {file.get_source_file_name()}  "
                        f"Content: {file.get_processed_file_data(max_words=max_words)}"
                    )
                    contents.append(text)
                elif file_format == FileFormatEnum.IMAGE:
                    mime_type = file.get_mime_type()
                    if mime_type not in ["image/jpeg", "image/png", "image/gif", "image/webp"]:
                        raise ValueError(f"Unsupported media type: {mime_type} for file {file.name}")

                    data_content = file.get_processed_file_data_content_only()
                    pcontent = f"data:{mime_type};base64,{data_content}"
                    contents.append(pcontent)
                else:
                    logger.error(
                        f"_convert_files_for_image_models: Unknown file_format {file_format} for file {file.name}"
                    )
            except Exception as e:
                logger.error(f"Error processing file {file.name}: {e}")

        return " ".join(contents)

    # -------------------------------------------------------------------------

    # Tools & Structured output
    @classmethod
    def convert_function_parameter(
        cls,
        param: FunctionParameter,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert FunctionParameter to OpenAI format"""
        result = param.model_dump(
            exclude={"required", "allowed_values", "default"},
        )
        return result

    @classmethod
    def convert_function_parameters(
        cls,
        params: FunctionParameters,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert FunctionParameters to OpenAI format"""
        # Create a new dictionary with transformed properties
        result = {
            "type": params.type,
            "properties": {name: cls.convert_function_parameter(param) for name, param in params.properties.items()},
        }

        # Auto-build the required list based on parameters marked as required
        required_params = [name for name, param in params.properties.items() if param.required]

        # Only include required field if there are required parameters
        if required_params:
            result["required"] = required_params
        elif params.required:  # If manually specified required array exists
            result["required"] = params.required

        return result

    @classmethod
    def convert_function_definition(
        cls,
        func_def: FunctionDefinition,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert FunctionDefinition to OpenAI format"""
        return {
            "name": func_def.name,
            "description": func_def.description,
            "parameters": cls.convert_function_parameters(func_def.parameters),
        }

    @classmethod
    def convert_tool(
        cls,
        tool: ToolDefinition,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> Any:
        """Convert ToolDefinition to OpenAI format"""
        return {
            "type": "function",
            "function": cls.convert_function_definition(tool.function),
        }

    @classmethod
    def convert_tool_choice(
        cls,
        tool_choice: ToolChoice,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> Any:
        """Convert ToolChoice to OpenAI format"""
        if tool_choice is None:
            return None

        if tool_choice.type is None:
            return None
        elif tool_choice.type == "zero_or_more":
            return "auto"
        elif tool_choice.type == "one_or_more":
            return "required"
        elif tool_choice.type == "specific":
            return {"type": "function", "name": tool_choice.specific_tool_name}

    @classmethod
    def convert_structured_output(
        cls,
        structured_output: StructuredOutputConfig,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert StructuredOutputConfig to OpenAI format"""
        # Get the original JSON schema from Pydantic or dict
        schema = structured_output.get_schema()

        # Ensure additionalProperties is set in the root schema and all nested definitions
        schema["additionalProperties"] = False

        # Also set additionalProperties for nested schemas
        if "$defs" in schema:
            for def_schema in schema["$defs"].values():
                def_schema["additionalProperties"] = False

        # Extract the name from the title and use it for schema name
        schema_name = schema.get("title", "output")

        # Clean up JSON Schema keywords that OpenAI doesn't permit
        if "properties" in schema:
            for prop in schema["properties"].values():
                # Remove numeric constraints that are not permitted. (from pydantic ge, le)
                prop.pop("minimum", None)
                prop.pop("maximum", None)

        # Return formatted response format
        return {
            "type": "json_schema",
            "json_schema": {
                "name": schema_name,
                "schema": schema,
                # "strict": True,
            },
        }



import logging

from google import genai
from google.genai.types import HttpOptions as GooogleHttpOptions

from dhenara.ai.providers.base import AIModelProviderClientBase
from dhenara.ai.providers.shared import APIProviderSharedFns
from dhenara.ai.types.genai.ai_model import AIModelAPIProviderEnum

from .formatter import GoogleFormatter

logger = logging.getLogger(__name__)


class GoogleAIClientBase(AIModelProviderClientBase):
    """Base class for all Google AI Clients"""

    formatter = GoogleFormatter

    def initialize(self) -> None:
        pass

    def cleanup(self) -> None:
        pass

    def _get_client_params(self, api) -> tuple[str, dict]:
        """Common logic for both sync and async clients"""

        # _http_pars= self._get_client_http_params(api)
        _http_pars = {}
        if self.config.timeout:
            _http_pars["http_options"] = GooogleHttpOptions(
                timeout=int(self.config.timeout) * 1000,  # In milli seconds
            )

        if api.provider == AIModelAPIProviderEnum.GOOGLE_AI:
            params = {
                "api_key": api.api_key,
                # "http_options": {"api_version": "v1beta"},
                **_http_pars,
            }
            return "google_ai", params
        elif api.provider == AIModelAPIProviderEnum.GOOGLE_VERTEX_AI:
            client_params = APIProviderSharedFns.get_vertex_ai_credentials(api)
            params = {
                "vertexai": True,
                "credentials": client_params["credentials"],
                "project": client_params["project_id"],
                "location": client_params["location"],
                # "http_options": {"api_version": "v1beta"},
                **_http_pars,
            }
            return "vertex_ai", params
        else:
            error_msg = f"Unsupported API provider {api.provider} for Google AI"
            logger.error(error_msg)
            raise ValueError(error_msg)

    def _setup_client_sync(self) -> genai.Client:
        """Get the appropriate sync Google AI client"""
        api = self.model_endpoint.api
        client_type, params = self._get_client_params(api)
        return genai.Client(**params)

    async def _setup_client_async(self) -> genai.Client:
        """Get the appropriate async Google AI client"""
        api = self.model_endpoint.api
        client_type, params = self._get_client_params(api)
        return genai.Client(**params).aio



from dhenara.ai.types.genai.ai_model import (
    AIModelFunctionalTypeEnum,
    AIModelProviderEnum,
    ChatModelCostData,
    ChatModelSettings,
    FoundationModel,
)

GPT4o = FoundationModel(
    model_name="gpt-4o",
    display_name="GPT-4o",
    provider=AIModelProviderEnum.OPEN_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=128000,
        max_output_tokens=16384,
    ),
    valid_options={},
    metadata={
        "details": "OpenAI GPT-4o model, optimized for conversational AI.",
    },
    order=10,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=2.5,
        output_token_cost_per_million=10.0,
    ),
)


GPT4oMini = FoundationModel(
    model_name="gpt-4o-mini",
    display_name="GPT-4o-mini",
    provider=AIModelProviderEnum.OPEN_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=128000,
        max_output_tokens=16384,
    ),
    valid_options={},
    metadata={
        "details": "OpenAI's affordable and intelligent small model for fast, lightweight tasks.",
    },
    order=11,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=0.15,
        output_token_cost_per_million=0.60,
    ),
)

O1 = FoundationModel(
    model_name="o1",
    display_name="o1",
    provider=AIModelProviderEnum.OPEN_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=200000,
        max_output_tokens=100000,
    ),
    valid_options={},
    metadata={
        "details": "OpenAI o1 model, optimized for reasoning.",
    },
    order=20,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=15.0,
        output_token_cost_per_million=60.0,
    ),
)

O1Mini = FoundationModel(
    model_name="o1-mini",
    display_name="o1-mini",
    provider=AIModelProviderEnum.OPEN_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=128000,
        max_output_tokens=65536,
    ),
    valid_options={},
    metadata={
        "details": "OpenAI o1-mini model, optimized for reasoning.",
        "display_order": 20,
    },
    order=21,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=1.10,
        output_token_cost_per_million=4.40,
    ),
)


O3 = FoundationModel(
    model_name="o3",
    display_name="o3",
    provider=AIModelProviderEnum.OPEN_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=200000,
        max_output_tokens=100000,
    ),
    valid_options={},
    metadata={
        "details": "OpenAI o3 model, optimized for reasoning.",
    },
    order=20,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=2.0,
        output_token_cost_per_million=8.0,
    ),
)


O3Mini = FoundationModel(
    model_name="o3-mini",
    display_name="o3-mini",
    provider=AIModelProviderEnum.OPEN_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=200000,
        max_output_tokens=100000,
    ),
    valid_options={},
    metadata={
        "details": "OpenAI o3-mini model, optimized for reasoning.",
    },
    order=22,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=1.10,
        output_token_cost_per_million=4.40,
    ),
)

O4Mini = FoundationModel(
    model_name="o4-mini",
    display_name="o4-mini",
    provider=AIModelProviderEnum.OPEN_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=200000,
        max_output_tokens=100000,
    ),
    valid_options={},
    metadata={
        "details": "OpenAI o4-mini model, optimized for reasoning.",
    },
    order=22,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=1.10,
        output_token_cost_per_million=4.40,
    ),
)


GPT41 = FoundationModel(
    model_name="gpt-4.1",
    display_name="GPT-4.1",
    provider=AIModelProviderEnum.OPEN_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=1047576,
        max_output_tokens=32768,
    ),
    valid_options={},
    metadata={
        "details": "Flagship model for complex tasks. It is well suited for problem solving across domains.",
    },
    order=11,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=2.0,
        output_token_cost_per_million=8.0,
    ),
)


GPT41Mini = FoundationModel(
    model_name="gpt-4.1-mini",
    display_name="GPT-4.1-mini",
    provider=AIModelProviderEnum.OPEN_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=1047576,
        max_output_tokens=32768,
    ),
    valid_options={},
    metadata={
        "details": "GPT-4.1 mini provides a balance between intelligence, speed, and cost.",
    },
    order=11,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=0.40,
        output_token_cost_per_million=1.60,
    ),
)


GPT41Nano = FoundationModel(
    model_name="gpt-4.1-nano",
    display_name="GPT-4.1-nano",
    provider=AIModelProviderEnum.OPEN_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=1047576,
        max_output_tokens=32768,
    ),
    valid_options={},
    metadata={
        "details": "GPT-4.1 nano is the fastest, most cost-effective GPT-4.1 model.",
    },
    order=11,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=0.10,
        output_token_cost_per_million=0.40,
    ),
)


CHAT_MODELS = [GPT41, GPT41Mini, GPT41Nano, GPT4o, GPT4oMini, O1, O1Mini, O3, O3Mini, O4Mini]



import json
import logging
import os
from pathlib import Path

import yaml
from pydantic import Field

from dhenara.ai.types.genai import MODEL_TO_API_MAPPING, PROVIDER_CONFIGS
from dhenara.ai.types.genai.ai_model import AIModel, AIModelAPI, AIModelEndpoint, FoundationModel
from dhenara.ai.types.genai.foundation_models import ALL_FOUNDATION_MODELS
from dhenara.ai.types.resource._resource_config_item import ResourceConfigItem, ResourceConfigItemTypeEnum
from dhenara.ai.types.shared.base import BaseModel

logger = logging.getLogger(__name__)


class ResourceConfig(BaseModel):
    """
    Configuration for AI resources including models, APIs, and endpoints.
    Manages loading credentials, initializing APIs, and creating model endpoints.
    """

    models: list[AIModel | FoundationModel] = Field(
        default_factory=list,
        description="AIModels",
    )
    model_apis: list[AIModelAPI] = Field(
        default_factory=list,
        description="AIModel APIs",
    )
    model_endpoints: list[AIModelEndpoint] = Field(
        default_factory=list,
        description="AIModel Endpoints",
    )

    def get_api(self, api_provider: str | None = None) -> AIModelEndpoint:
        return next((api for api in self.model_apis if api.provider == api_provider), None)

    def get_model_endpoint(self, model_name: str, api_provider: str | None = None) -> AIModelEndpoint:
        """
        Retrieves an endpoint by model name.

        Args:
            model_name: Model name

        Returns:
            AIModelEndpoint instance
        """

        query = {"model_name": model_name}
        if api_provider:
            query["api_provider"] = api_provider

        return self.get_resource(
            ResourceConfigItem(
                item_type=ResourceConfigItemTypeEnum.ai_model_endpoint,
                query=query,
            )
        )

    def get_resource(self, resource_item: ResourceConfigItem) -> AIModelEndpoint:  # |RagEndpoint
        """
        Retrieves a resource based on the resource specification.

        Args:
            resource_item: ResourceConfigItem model instance

        Returns:
            AIModelEndpoint instance

        Raises:
            ValueError: If object not found or query invalid
        """
        if not resource_item.query:
            raise ValueError("Query must be provided")

        try:
            if resource_item.item_type == ResourceConfigItemTypeEnum.ai_model_endpoint:
                # Filter based on query parameters
                for endpoint in self.model_endpoints:
                    matches = True

                    for key, value in resource_item.query.items():
                        if key == "reference_number" and endpoint.reference_number != value:
                            matches = False
                            break
                        elif key == "model_name" and endpoint.ai_model.model_name != value:
                            matches = False
                            break
                        elif key == "model_display_name" and endpoint.ai_model.display_name != value:
                            matches = False
                            break
                        elif key == "api_provider" and endpoint.api.provider != value:
                            matches = False
                            break

                    if matches:
                        return endpoint

                # Create query description for error message
                query_desc = ", ".join(f"{k}={v}" for k, v in resource_item.query.items())
                logger.error(f"No endpoint found matching query: {query_desc}")
                return None

            else:
                logger.error(f"Unsupported resource type: {resource_item.item_type}")
                return None

        except Exception as e:
            logger.error(f"Error fetching resource: {e}")
            return None

    def load_from_file(
        self,
        credentials_file: str,
        models: list[AIModel] | None = None,
        init_endpoints: bool = False,
    ):
        """
        Initialize the ResourceConfig with credentials from a file and optional overrides.

        Args:
            credentials_file: Path to the credentials file (JSON or YAML)
            models: Optional list of models to override default ALL_FOUNDATION_MODELS
            mapping_override: Optional dictionary to override default model-to-API mappings
        """

        # Load credentials
        credentials = self._load_credentials_from_file(credentials_file)

        # Initialize APIs with loaded credentials
        self._initialize_apis(credentials)

        if models:
            self.models = models

        if init_endpoints:
            if not self.models:
                self.models = [*ALL_FOUNDATION_MODELS]

            # Create endpoints from models and APIs
            self._initialize_endpoints()

    def _load_credentials_from_file(self, credentials_file: str) -> dict:
        """
        Load credentials from a file (JSON or YAML format).

        Args:
            credentials_file: Path to the credentials file

        Returns:
            Dictionary containing credentials for each provider

        Raises:
            FileNotFoundError: If the credentials file doesn't exist
            ValueError: If the file format is unsupported
        """
        # Expand the tilde to home directory if present
        expanded_path = os.path.expanduser(credentials_file)
        path = Path(expanded_path)

        if not path.exists():
            raise FileNotFoundError(f"Credentials file not found: {credentials_file}")

        ext = path.suffix.lower()
        with open(path) as f:
            if ext == ".json":
                credentials = json.load(f)
            elif ext in [".yaml", ".yml"]:
                credentials = yaml.safe_load(f)
            else:
                raise ValueError(f"Unsupported file format: {ext}. Use .json, .yaml, or .yml")

        # Validate loaded credentials
        return credentials

    def _initialize_apis(self, credentials: dict) -> None:
        """
        Initialize API clients based on loaded credentials.

        Args:
            credentials: Dictionary of credentials by provider
        """
        for provider_str, provider_creds_data in credentials.items():
            try:
                new_api = AIModelAPI(
                    provider=provider_str,
                    api_key=provider_creds_data.get("api_key", None),
                    credentials=provider_creds_data.get("credentials", {}),
                    config=provider_creds_data.get("config", {}),
                )

                # If validation passes, add the endpint
                self.model_apis.append(new_api)

            except Exception as e:  # noqa: PERF203
                logger.exception(f"Error initializing API for provider '{provider_str}': {e}")

    def _initialize_endpoints(self) -> None:
        """
        Create model endpoints by matching models with compatible APIs.
        For each model, finds a compatible API and creates an endpoint.
        """
        # Track created endpoints to avoid duplicates
        created_endpoints = set()

        for model in self.models:
            # Get compatible API providers for this model's provider
            compatible_providers = MODEL_TO_API_MAPPING.get(model.provider, [])

            # Find the first available API that can handle this model
            for api_provider in compatible_providers:
                api = next((api for api in self.model_apis if api.provider == api_provider), None)

                if api:
                    # Create a unique identifier for this model-API combination
                    endpoint_id = f"{model.model_name}_{api.provider}"

                    # Skip if already created
                    if endpoint_id in created_endpoints:
                        continue

                    # Create the endpoint
                    endpoint = AIModelEndpoint(
                        reference_number=endpoint_id,
                        ai_model=model,
                        api=api,
                    )

                    self.model_endpoints.append(endpoint)
                    created_endpoints.add(endpoint_id)

                    # Once we've created an endpoint for this model, move to the next model
                    break

    @classmethod
    def create_credentials_template(cls, output_file: str = "credentials.yaml") -> None:
        """
        Create a template credentials file with required fields for each provider.

        Args:
            output_file: Path to save the template file
        """
        # Start with a header comment
        header = (
            "# Dhenara AI Provider Credentials\n"
            "# Replace placeholder values with your actual API keys and remove unused items\n\n"
        )

        template = {}

        for provider_enum, config in PROVIDER_CONFIGS.items():
            provider_str = provider_enum.value
            template[provider_str] = {}

            # Add API key if required
            if config.api_key_required:
                template[provider_str]["api_key"] = f"<YOUR_{provider_str.upper()}_API_KEY>"

            # Add required credential fields
            for field_config in config.credentials_required_fields:
                field_name = field_config.field_name
                template[provider_str][field_name] = f"<YOUR_{provider_str.upper()}_{field_name.upper()}>"

            for field_config in config.config_required_fields:
                field_name = field_config.field_name
                template[provider_str][field_name] = f"<YOUR_{provider_str.upper()}_{field_name.upper()}>"

            # Do not add the optional fields directly to the template dictionary with the comment
            # Instead, we'll handle comments specially during YAML dumping

        # Write the template to a file
        ext = os.path.splitext(output_file)[1].lower()

        if ext == ".json":
            with open(output_file, "w") as f:
                json.dump(template, f, indent=2)
        else:
            # For YAML, we need to create the content manually to include comments properly
            yaml_content = header

            for provider_enum, config in PROVIDER_CONFIGS.items():
                provider_str = provider_enum.value
                yaml_content += f"{provider_str}:\n"

                # Add API key if required
                if config.api_key_required:
                    yaml_content += f"  api_key: <YOUR_{provider_str.upper()}_API_KEY>\n"

                # credentials
                if config.credentials_required_fields or config.credentials_optional_fields:
                    yaml_content += "  credentials:\n"

                # Add required credential fields
                for field_config in config.credentials_required_fields:
                    field_name = field_config.field_name
                    yaml_content += f"    {field_name}: <YOUR_{provider_str.upper()}_{field_name.upper()}>\n"

                # Add optional credential fields with comment
                for field_config in config.credentials_optional_fields:
                    field_name = field_config.field_name
                    yaml_content += f"    {field_name}: <YOUR_{provider_str.upper()}_{field_name.upper()}> # Optional\n"

                # config
                if config.config_required_fields or config.config_optional_fields:
                    yaml_content += "  config:\n"

                # Add required config fields
                for field_config in config.config_required_fields:
                    field_name = field_config.field_name
                    yaml_content += f"    {field_name}: <YOUR_{provider_str.upper()}_{field_name.upper()}>\n"

                # Add optional config fields with comment
                for field_config in config.config_optional_fields:
                    field_name = field_config.field_name
                    yaml_content += f"    {field_name}: <YOUR_{provider_str.upper()}_{field_name.upper()}> # Optional\n"

                # Add an extra newline between providers
                yaml_content += "\n"

            # Write the YAML content to file
            with open(output_file, "w") as f:
                f.write(yaml_content)

        logger.info(f"Created credentials template at {output_file}")
        logger.info("Edit this file with your API credentials before loading")



import logging
from typing import Any

from dhenara.ai.providers.base import BaseFormatter
from dhenara.ai.types.genai.ai_model import AIModelEndpoint
from dhenara.ai.types.genai.dhenara.request import (
    FunctionDefinition,
    FunctionParameter,
    FunctionParameters,
    PromptMessageRoleEnum,
    StructuredOutputConfig,
    ToolChoice,
    ToolDefinition,
)
from dhenara.ai.types.genai.dhenara.request.data import FormattedPrompt
from dhenara.ai.types.shared.file import FileFormatEnum, GenericFile

logger = logging.getLogger(__name__)


class GoogleFormatter(BaseFormatter):
    """
    Formatter for converting Dhenara types to Google-specific formats and vice versa.
    """

    role_map = {
        PromptMessageRoleEnum.USER: "user",
        PromptMessageRoleEnum.ASSISTANT: "model",
        PromptMessageRoleEnum.SYSTEM: "system",  # NOTE:  Don't care as system instructions are taken care separately
    }

    @classmethod
    def convert_prompt(
        cls,
        formatted_prompt: FormattedPrompt,
        model_endpoint: AIModelEndpoint | None = None,
        files: list[GenericFile] | None = None,
        max_words_file: int | None = None,
    ) -> dict[str, Any]:
        # Map Dhenara formats to provider format
        parts = []
        file_contents = None
        if files:
            file_contents = cls.convert_files_to_provider_content(
                files=files,
                model_endpoint=model_endpoint,
                max_words=max_words_file,
            )

        # if model_endpoint.ai_model.functional_type == AIModelFunctionalTypeEnum.IMAGE_GENERATION:
        #    return cls._convert_image_model_prompt(
        #        formatted_prompt=formatted_prompt,
        #        model_endpoint=model_endpoint,
        #        file_contents=file_contents,
        #    )

        parts.append(
            {
                "text": formatted_prompt.text,
            }
        )
        if file_contents:
            parts.extend(file_contents)

        role = cls.role_map.get(formatted_prompt.role)
        return {"role": role, "parts": parts}

    @classmethod
    def convert_instruction_prompt(
        cls,
        formatted_prompt: FormattedPrompt,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        # There is no native support for `system` prompt. So set role always send as `user` role,
        # so that beta models can send them as prompt.
        # For other models, the text will be send as seperate argument
        role = cls.role_map.get(PromptMessageRoleEnum.USER)

        return {
            "role": role,
            "parts": [
                {
                    "text": formatted_prompt.text,
                }
            ],
        }

    @classmethod
    def convert_files_to_provider_content(
        cls,
        files: list[GenericFile],
        model_endpoint: AIModelEndpoint | None = None,
        max_words: int | None = None,
    ) -> list[dict[str, Any]]:
        # if model_endpoint.ai_model.functional_type == AIModelFunctionalTypeEnum.IMAGE_GENERATION:
        #    return cls._convert_files_for_image_models(
        #        files=files,
        #        model_endpoint=model_endpoint,
        #        max_words=max_words,
        #    )

        contents = []
        for file in files:
            file_format = file.get_file_format()
            if file_format in [FileFormatEnum.COMPRESSED, FileFormatEnum.TEXT]:
                text = f"\nFile: {file.get_source_file_name()}  Content: {file.get_processed_file_data(max_words)}"
                pcontent = text
                contents.append(
                    {
                        "text": pcontent,
                    }
                )
            elif file_format in [FileFormatEnum.IMAGE]:
                mime_type = file.get_mime_type()
                contents.append(
                    {
                        "inline_data": {
                            "data": file.get_processed_file_data_content_only(),  # Bytes type
                            "mime_type": mime_type,
                        },
                    }
                )
            else:
                logger.error(f"get_prompt_file_contents: Unknown file_format {file_format} for file {file.name} ")

        return contents

    # -------------------------------------------------------------------------

    # Tools & Structured output
    @classmethod
    def convert_function_parameter(
        cls,
        param: FunctionParameter,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert FunctionParameter to Google format"""
        result = param.model_dump(
            exclude={"required", "allowed_values", "default"},
        )
        return result

    @classmethod
    def convert_function_parameters(
        cls,
        params: FunctionParameters,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert FunctionParameters to Google format"""
        # Create a new dictionary with transformed properties
        result = {
            "type": params.type,
            "properties": {name: cls.convert_function_parameter(param) for name, param in params.properties.items()},
        }

        # Auto-build the required list based on parameters marked as required
        required_params = [name for name, param in params.properties.items() if param.required]

        # Only include required field if there are required parameters
        if required_params:
            result["required"] = required_params
        elif params.required:  # If manually specified required array exists
            result["required"] = params.required

        return result

    @classmethod
    def convert_function_definition(
        cls,
        func_def: FunctionDefinition,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert FunctionDefinition to Google format"""
        return {
            "name": func_def.name,
            "description": func_def.description,
            "parameters": cls.convert_function_parameters(func_def.parameters),
        }

    @classmethod
    def convert_tool(
        cls,
        tool: ToolDefinition,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> Any:
        """Convert ToolDefinition to Google format"""

        return {"function_declarations": [cls.convert_function_definition(tool.function)]}

    @classmethod
    def convert_tool_choice(
        cls,
        tool_choice: ToolChoice,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> Any:
        """Convert ToolChoice to Google format"""

        if tool_choice.type is None:
            return None
        elif tool_choice.type == "zero_or_more":
            _cfg = {
                "mode": "AUTO",
            }
        elif tool_choice.type == "one_or_more":
            _cfg = {
                "mode": "ANY",
            }
        elif tool_choice.type == "specific":
            _cfg = {
                "mode": "AUTO",
                "allowed_function_names": [tool_choice.specific_tool_name],
            }

        return {"function_calling_config": _cfg}

    @classmethod
    def convert_structured_output(
        cls,
        structured_output: StructuredOutputConfig,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert StructuredOutputConfig to Google format"""
        # return structured_output.output_schema

        def _clean_schema_for_api(schema):
            """Remove `additionalProperties` in nested list/dict"""
            if isinstance(schema, dict):
                # Remove additionalProperties: false
                if "additionalProperties" in schema and schema["additionalProperties"] is False:
                    del schema["additionalProperties"]

                # Remove empty examples lists
                if "examples" in schema and not schema["examples"]:
                    del schema["examples"]

                # Process nested schemas
                for _key, value in list(schema.items()):
                    if isinstance(value, (dict, list)):
                        _clean_schema_for_api(value)

            elif isinstance(schema, list):
                for item in schema:
                    if isinstance(item, (dict, list)):
                        _clean_schema_for_api(item)

            return schema

        # Get the original JSON schema from Pydantic or dict
        _schema = structured_output.get_schema()
        return _clean_schema_for_api(_schema)



from dhenara.ai.types.genai.ai_model import (
    AIModelProviderEnum,
    ImageResponseUsage,
    UsageCharge,
)
from dhenara.ai.types.genai.dhenara.request import PromptMessageRoleEnum
from dhenara.ai.types.genai.dhenara.request.data import Prompt, PromptConfig
from dhenara.ai.types.shared.base import BaseModel

from ._content_items._image_items import ImageResponseContentItem
from ._metadata import AIModelCallResponseMetaData


class ImageResponseChoice(BaseModel):
    """A single image generation choice/result"""

    index: int
    contents: list[ImageResponseContentItem] = []

    class Config:
        json_schema_extra = {
            "example": {
                "index": 0,
                "content": {
                    "content_format": "url",
                    "content_url": "https://api.example.com/images/123.jpg",
                },
            }
        }


class ImageResponse(BaseModel):
    """Complete response from an AI image generation model

    Contains the generated images, usage information, and provider-specific metadata
    """

    model: str
    provider: AIModelProviderEnum
    usage: ImageResponseUsage | None
    usage_charge: UsageCharge | None
    choices: list[ImageResponseChoice]
    metadata: AIModelCallResponseMetaData | dict = {}

    def to_prompt(
        self,
        choice_index: int = 0,
        max_words_text: int | None = None,
    ) -> "Prompt":
        """Convert response to a context message for next turn"""

        # Get text from the first choice's contents
        if not self.choices:
            return None

        choice = self.choices[choice_index]
        if not choice.contents:
            return None

        # TODO
        # Process generated images if availae and add as files
        image_files = []

        # Create and return Prompt object
        return Prompt(
            role=PromptMessageRoleEnum.ASSISTANT,
            text="Generated Image",
            files=image_files,
            config=PromptConfig(
                max_words_text=max_words_text,
                max_words_file=None,
            ),
        )

    def preview_dict(self):
        """
        Returns a preview version of the response excluding the full content of choices
        but including metadata about them
        """
        _dict = self.model_dump(exclude=["choices"])

        # Add summary information about choices instead of full content
        choice_summaries = []
        for choice in self.choices:
            choice_summary = {
                "index": choice.index,
                "content_count": len(choice.contents),
                "contents_summary": [
                    {
                        "index": content.index,
                        "content_format": content.content_format,
                    }
                    for content in choice.contents
                ],
            }
            choice_summaries.append(choice_summary)

        _dict["choices_summary"] = choice_summaries
        return _dict



from pydantic import Field, model_validator

from dhenara.ai.types.genai.ai_model import (
    AIModelAPI,
    AIModelAPIProviderEnum,
    AIModelProviderEnum,
    BaseAIModel,
    ChatModelCostData,
    ChatResponseUsage,
    ImageModelCostData,
    ImageResponseUsage,
)
from dhenara.ai.types.shared.base import BaseModel


class AIModelEndpoint(BaseModel):
    """
    Pydantic model representing an AI model endpoint configuration.
    """

    api: AIModelAPI = Field(
        ...,
        description="Reference to API credentials",
    )
    ai_model: BaseAIModel = Field(
        ...,
        description="Reference to AI model",
    )
    order: int = Field(
        0,
        description="Order for display purposes",
    )
    enabled: bool = Field(
        True,  # noqa: FBT003
        description="Whether the endpoint is enabled",
    )
    cost_data: ChatModelCostData | ImageModelCostData | None = Field(
        None,
        description="Matching foundation model for parameter preloading",
    )
    reference_number: str | None = Field(
        None,
        description="reference number. Should be unique if not None",
    )

    @model_validator(mode="after")
    def _validate_cost_data(self) -> "AIModelEndpoint":
        if self.cost_data:
            (setting_model, cost_model) = BaseAIModel.get_pydantic_model_classes(self.functional_type)
            if not isinstance(self.cost_data, cost_model):
                raise ValueError(f"For {self.functional_type} endpoins, cost data must be type {cost_model} or None.")

        return self

    def get_cost_data(self):
        if self.cost_data:
            return self.cost_data
        else:
            return self.ai_model.get_cost_data()

    def calculate_usage_charge(
        self,
        usage: ChatResponseUsage | ImageResponseUsage,
    ):
        cost_data = self.get_cost_data()
        return cost_data.calculate_usage_charge(usage)


# Default mapping of model providers to API providers
MODEL_TO_API_MAPPING = {
    # Anthropic models
    AIModelProviderEnum.ANTHROPIC: [
        AIModelAPIProviderEnum.ANTHROPIC,
        AIModelAPIProviderEnum.AMAZON_BEDROCK,
        AIModelAPIProviderEnum.GOOGLE_VERTEX_AI,
    ],
    AIModelProviderEnum.OPEN_AI: [
        AIModelAPIProviderEnum.OPEN_AI,
        AIModelAPIProviderEnum.MICROSOFT_OPENAI,
    ],
    AIModelProviderEnum.GOOGLE_AI: [
        AIModelAPIProviderEnum.GOOGLE_VERTEX_AI,  # If both are available, prioritize vertext_ai
        AIModelAPIProviderEnum.GOOGLE_AI,
    ],
    AIModelProviderEnum.DEEPSEEK: [
        AIModelAPIProviderEnum.MICROSOFT_AZURE_AI,
    ],
}



from dhenara.ai.types.genai.ai_model import (
    AIModelFunctionalTypeEnum,
    AIModelProviderEnum,
    ChatModelCostData,
    ChatModelSettings,
    FoundationModel,
)

Claude40Sonnet = FoundationModel(
    model_name="claude-sonnet-4-0",
    display_name="Claude Sonnet 4",
    provider=AIModelProviderEnum.ANTHROPIC,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=200000,
        max_output_tokens=8192,
        supports_reasoning=True,
        max_reasoning_tokens=32000,
        max_output_tokens_reasoning_mode=64000,
    ),
    valid_options={},
    metadata={
        # "version_suffix": "-latest",
    },
    order=70,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=3.0,
        output_token_cost_per_million=15.0,
    ),
)

Claude40Opus = FoundationModel(
    model_name="claude-opus-4-0",
    display_name="Claude Opus 4",
    provider=AIModelProviderEnum.ANTHROPIC,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=200000,
        max_output_tokens=8192,
        supports_reasoning=True,
        max_reasoning_tokens=32000,
        max_output_tokens_reasoning_mode=32000,
    ),
    valid_options={},
    metadata={
        # "version_suffix": "-latest",
    },
    order=75,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=15.0,
        output_token_cost_per_million=75.0,
    ),
)

Claude37Sonnet = FoundationModel(
    model_name="claude-3-7-sonnet",
    display_name="Claude Sonnet 3.7",
    provider=AIModelProviderEnum.ANTHROPIC,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=200000,
        max_output_tokens=8192,
        supports_reasoning=True,
        max_reasoning_tokens=32000,
        max_output_tokens_reasoning_mode=64000,
    ),
    valid_options={},
    metadata={
        "details": "Model, with highest level of intelligence and capability.",
        "version_suffix": "-latest",  # NOTE: Version is required for Anthropic API calls
    },
    order=81,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=3.0,
        output_token_cost_per_million=15.0,
    ),
)


Claude35Sonnet = FoundationModel(
    model_name="claude-3-5-sonnet",
    display_name="Claude Sonnet 3.5",
    provider=AIModelProviderEnum.ANTHROPIC,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=200000,
        max_output_tokens=8192,
    ),
    valid_options={},
    metadata={
        "details": "Model, with highest level of intelligence and capability.",
        "version_suffix": "-latest",  # NOTE: Version is required for Anthropic API calls
    },
    order=81,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=3.0,
        output_token_cost_per_million=15.0,
    ),
)

Claude35Haiku = FoundationModel(
    model_name="claude-3-5-haiku",
    display_name="Claude Haiku 3.5",
    provider=AIModelProviderEnum.ANTHROPIC,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=200000,
        max_output_tokens=8192,
    ),
    valid_options={},
    metadata={
        "details": "Fastest, most cost-effective model.",
        "version_suffix": "-latest",  # NOTE: Version is required for Anthropic API calls
    },
    order=82,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=1.0,
        output_token_cost_per_million=5.0,
    ),
)


Claude3Opus = FoundationModel(
    model_name="claude-3-opus",
    display_name="Claude 3 Opus",
    provider=AIModelProviderEnum.ANTHROPIC,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=200000,
        max_output_tokens=4096,
    ),
    valid_options={},
    metadata={
        "details": "Powerful model for highly complex tasks",
        "version_suffix": "-latest",  # NOTE: Version is required for Anthropic API calls
    },
    order=93,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=15.0,
        output_token_cost_per_million=75.0,
    ),
)

CHAT_MODELS = [
    Claude40Opus,
    Claude40Sonnet,
    Claude37Sonnet,
    Claude35Sonnet,
    Claude35Haiku,
    Claude3Opus,
]



import asyncio
import concurrent.futures
import functools
import logging
import threading
from collections.abc import Callable
from dataclasses import dataclass
from typing import ParamSpec, TypeVar

logger = logging.getLogger(__name__)

P = ParamSpec("P")
T = TypeVar("T")


@dataclass
class SyncWrapperConfig:
    max_workers: int = 10
    default_timeout: float | None = None
    thread_name_prefix: str = "async_wrapper_"
    shutdown_timeout: float = 60.0


class AsyncToSyncWrapper:
    """
    Handles conversion of async functions to sync functions with proper
    thread and event loop management.
    """

    def __init__(self, config: SyncWrapperConfig | None = None):
        self._config = config or SyncWrapperConfig()
        self._executor: concurrent.futures.ThreadPoolExecutor | None = None
        self._lock = threading.Lock()

    def __call__(self, async_func: Callable[P, T]) -> Callable[P, T]:
        @functools.wraps(async_func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                loop = None

            if loop and loop.is_running():
                return self._run_in_new_loop(async_func, *args, **kwargs)
            return self._run_async_func(async_func, *args, **kwargs)

        return wrapper

    def _get_executor(self) -> concurrent.futures.ThreadPoolExecutor:
        with self._lock:
            if self._executor is None:
                self._executor = concurrent.futures.ThreadPoolExecutor(
                    max_workers=self._config.max_workers,
                    thread_name_prefix=self._config.thread_name_prefix,
                )
            return self._executor

    def _run_in_new_loop(
        self,
        async_func,
        *args,
        timeout: float | None = None,
        **kwargs,
    ):
        future = concurrent.futures.Future()
        timeout = timeout or self._config.default_timeout

        def run_in_thread():
            try:
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                try:
                    result = loop.run_until_complete(async_func(*args, **kwargs))
                    future.set_result(result)
                finally:
                    loop.close()
            except Exception as e:
                logger.exception("Error in async execution")
                future.set_exception(e)

        self._get_executor().submit(run_in_thread)
        return future.result(timeout=timeout)

    def _run_async_func(self, async_func, *args, **kwargs):
        return asyncio.run(async_func(*args, **kwargs))

    def cleanup(self):
        with self._lock:
            if self._executor:
                try:
                    self._executor.shutdown(wait=True, timeout=self._config.shutdown_timeout)
                except Exception:
                    logger.exception("Error during executor shutdown")
                finally:
                    self._executor = None


sync_wrapper = AsyncToSyncWrapper()
# def sync_wrapper(async_func):
#    """Decorator to convert async functions to sync functions"""
#
#    @wraps(async_func)
#    def sync_func(*args, **kwargs):
#        return asyncio.run(async_func(*args, **kwargs))
#
#    return sync_func



# ruff: noqa: E501

from dhenara.ai.types.genai.ai_model import (
    FoundationModel,
)

from .constants import ALL_FOUNDATION_MODELS


class FoundationModelFns:
    @staticmethod
    def get_foundation_model(name, all_models: list[FoundationModel] | None = None):
        _all_models = all_models if all_models else ALL_FOUNDATION_MODELS
        try:
            return next(model for model in _all_models if model.model_name == name)
        except StopIteration:
            return None



import logging

from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI

from dhenara.ai.providers.base import AIModelProviderClientBase
from dhenara.ai.types.genai.ai_model import AIModelAPIProviderEnum

from .formatter import OpenAIFormatter

logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
class OpenAIClientBase(AIModelProviderClientBase):
    """Base class for all OpenAI Clients"""

    formatter = OpenAIFormatter

    def initialize(self) -> None:
        pass

    def cleanup(self) -> None:
        pass

    def _get_client_params(self, api) -> tuple[str, dict]:
        """Common logic for both sync and async clients"""
        if api.provider == AIModelAPIProviderEnum.OPEN_AI:
            params = {
                "api_key": api.api_key,
                **self._get_client_http_params(api),
            }
            return "openai", params

        elif api.provider == AIModelAPIProviderEnum.MICROSOFT_OPENAI:
            client_params = api.get_provider_credentials()
            params = {
                "api_key": client_params["api_key"],
                "azure_endpoint": client_params["azure_endpoint"],
                "api_version": client_params["api_version"],
                **self._get_client_http_params(api),
            }
            return "azure_openai", params

        elif api.provider == AIModelAPIProviderEnum.MICROSOFT_AZURE_AI:
            client_params = api.get_provider_credentials()
            params = {
                "endpoint": client_params["azure_endpoint"],
                "credential": client_params["api_key"],
                **self._get_client_http_params(api),
            }
            return "azure_ai", params

        error_msg = f"Unsupported API provider {api.provider} for OpenAI functions"
        logger.error(error_msg)
        raise ValueError(error_msg)

    def _setup_client_sync(self) -> OpenAI | AzureOpenAI:
        """Get the appropriate sync OpenAI client"""
        api = self.model_endpoint.api
        client_type, params = self._get_client_params(api)

        if client_type == "openai":
            return OpenAI(**params)
        elif client_type == "azure_openai":
            return AzureOpenAI(**params)
        else:  # azure_ai
            from azure.ai.inference import ChatCompletionsClient
            from azure.core.credentials import AzureKeyCredential

            return ChatCompletionsClient(
                endpoint=params["endpoint"],
                credential=AzureKeyCredential(key=params["credential"]),
            )

    async def _setup_client_async(self) -> AsyncOpenAI | AsyncAzureOpenAI:
        """Get the appropriate async OpenAI client"""
        api = self.model_endpoint.api
        client_type, params = self._get_client_params(api)

        if client_type == "openai":
            return AsyncOpenAI(**params)
        elif client_type == "azure_openai":
            return AsyncAzureOpenAI(**params)
        else:  # azure_ai
            from azure.ai.inference.aio import ChatCompletionsClient
            from azure.core.credentials import AzureKeyCredential

            return ChatCompletionsClient(
                endpoint=params["endpoint"],
                credential=AzureKeyCredential(key=params["credential"]),
            )
